"""The build module contains helper classes for building from git repositories
via :py:class:`GitRepositoryBuild` and to perform multistage containerfile
builds via :py:class:`MultiStageBuild`.

"""

import tempfile
from dataclasses import dataclass
from os.path import basename
from os.path import join
from pathlib import Path
from string import Template
from subprocess import check_output
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from uuid import uuid4

from _pytest.config import Config
from _pytest.mark.structures import ParameterSet

from pytest_container.container import Container
from pytest_container.container import DerivedContainer
from pytest_container.container import container_and_marks_from_pytest_param
from pytest_container.logging import _logger
from pytest_container.runtime import OciRuntimeBase
from pytest_container.runtime import ToParamMixin


@dataclass(frozen=True)
class GitRepositoryBuild(ToParamMixin):
    """Test information storage for running builds using an external git
    repository. It is a required parameter for the `container_git_clone` and
    `host_git_clone` fixtures.
    """

    #: url of the git repository, can end with .git
    repository_url: str = ""
    #: an optional tag at which the repository should be checked out instead of
    #: using the default branch
    repository_tag: Optional[str] = None

    #: The command to run a "build" of the git repository inside a working
    #: copy.
    #: It can be left empty on purpose.
    build_command: str = ""

    def __post_init__(self) -> None:
        if not self.repository_url:
            raise ValueError("A repository url must be provided")

    def __str__(self) -> str:
        return self.repo_name

    @property
    def repo_name(self) -> str:
        """Name of the directory to which the repository will be checked out"""
        repo_without_dot_git = self.repository_url.replace(".git", "")
        repo_without_trailing_slash = (
            repo_without_dot_git[0:-1]
            if repo_without_dot_git[-1] == "/"
            else repo_without_dot_git
        )
        return basename(repo_without_trailing_slash)

    @property
    def clone_command(self) -> str:
        """Command to clone the repository at the appropriate tag"""
        clone_cmd_parts = ["git clone --depth 1"]
        if self.repository_tag:
            clone_cmd_parts.append(f"--branch {self.repository_tag}")
        clone_cmd_parts.append(self.repository_url)

        return " ".join(clone_cmd_parts)

    @property
    def test_command(self) -> str:
        """The full test command, including build_command and a cd into the
        correct folder.
        """
        cd_cmd = f"cd {self.repo_name}"
        if self.build_command:
            return f"{cd_cmd} && {self.build_command}"
        return cd_cmd


@dataclass
class MultiStageBuild:
    """Helper class to perform multi-stage container builds using the
    :py:class:`~pytest_container.container.Container` and
    :py:class:`~pytest_container.container.DerivedContainer` classes.

    This class is essentially just a very simple helper that will replace all
    variables in :py:attr:`containerfile_template` with the correct container
    ids, urls or names from the containers in :py:attr:`containers`.

    For example the following class:

    .. code-block:: python

       MultiStageBuild(
           containers={
               "builder": Container(url="registry.opensuse.org/opensuse/busybox:latest"),
               "runner1": "docker.io/alpine",
           },
           containerfile_template=r'''FROM $builder as builder

       FROM $runner1 as runner1
       ''',
       )

    would yield the following :file:`Containerfile`:

    .. code-block:: Dockerfile

       FROM registry.opensuse.org/opensuse/busybox:latest as builder

       FROM docker.io/alpine as runner1


    The resulting object can either be used to retrieve the rendered
    :file:`Containerfile` or to build the containers:

    .. code-block:: python

       id_of_runner1 = MULTI_STAGE_BUILD.build(
           tmp_path, pytestconfig, container_runtime, "runner1"
       )

    Where ``tmp_path`` and ``pytestconfig`` are the pytest fixtures and
    ``container_runtime`` is an instance of a child class of
    :py:class:`~pytest_container.runtime.OciRuntimeBase`. For further details,
    see :py:meth:`build`.

    """

    #: Template string of a :file:`Containerfile` where all containers from
    #: :py:attr:`containers` are inserted when retrieved via
    #: :py:attr:`containerfile`.
    containerfile_template: str

    #: A dictionary mapping the container names used in
    #: :py:attr:`containerfile_template` to
    #: :py:class:`~pytest_container.container.Container` or
    #: :py:class:`~pytest_container.container.DerivedContainer` objects or
    #: strings or any of the previous classes wrapped inside a `pytest.param
    #: <https://docs.pytest.org/en/stable/reference.html?#pytest.param>`_.
    containers: Dict[
        str, Union[Container, DerivedContainer, str, ParameterSet]
    ]

    @property
    def containerfile(self) -> str:
        """The rendered :file:`Containerfile` from the template supplied in
        :py:attr:`containerfile_template`.

        """
        return Template(self.containerfile_template).substitute(
            **{
                k: v
                if isinstance(v, str)
                else str(
                    container_and_marks_from_pytest_param(v)[0]._build_tag
                )
                for k, v in self.containers.items()
            }
        )

    def prepare_build(
        self,
        tmp_path: Path,
        container_runtime: OciRuntimeBase,
        rootdir: Path,
        extra_build_args: Optional[List[str]] = None,
    ) -> None:
        """Prepares the multistage build: it writes the rendered :file:`Containerfile`
        into ``tmp_path`` and prepares all containers in :py:attr:`containers` in
        the given ``rootdir``. Optional additional build arguments can be passed
        to the preparation of the containers

        """
        _logger.debug("Preparing multistage build")
        for _, container in self.containers.items():
            if not isinstance(container, str):
                container_and_marks_from_pytest_param(container)[
                    0
                ].prepare_container(
                    container_runtime, rootdir, extra_build_args
                )

        dockerfile_dest = tmp_path / "Dockerfile"
        with open(dockerfile_dest, "w", encoding="utf-8") as containerfile:
            _logger.debug(
                "Writing the following dockerfile into %s: %s",
                dockerfile_dest,
                self.containerfile,
            )
            containerfile.write(self.containerfile)

    @staticmethod
    def run_build_step(
        tmp_path: Path,
        runtime: OciRuntimeBase,
        target: Optional[str] = None,
        extra_build_args: Optional[List[str]] = None,
    ) -> str:
        """Run the multistage build in the given ``tmp_path`` using the supplied
        ``runtime``. This function requires :py:meth:`prepare_build` to be run
        beforehand.

        Args:
            tmp_path: the path in which the build was prepared.
            runtime: the container runtime which will be used to perform the
                build
            target: an optional target to which the build will be run, see `the
                upstream documentation
                <https://docs.docker.com/develop/develop-images/multistage-build/#stop-at-a-specific-build-stage>`_
                for more information

        Returns:
            Id of the final container that has been built
        """
        # This is an ugly, duplication of the launcher code
        with tempfile.TemporaryDirectory() as tmp_dir:
            iidfile = join(tmp_dir, str(uuid4()))
            cmd = (
                runtime.build_command
                + (extra_build_args or [])
                + [f"--iidfile={iidfile}"]
                + (["--target", target] if target else [])
                + [str(tmp_path)]
            )
            _logger.debug("Running multistage container build: %s", cmd)
            check_output(cmd)
            return runtime.get_image_id_from_iidfile(iidfile)

    def build(
        self,
        tmp_path: Path,
        rootdir_or_pytestconfig: Union[Path, Config],
        runtime: OciRuntimeBase,
        target: Optional[str] = None,
        extra_build_args: Optional[List[str]] = None,
    ) -> str:
        """Perform the complete multistage build to an optional target.

        Args:
            tmp_path: temporary directory into which the :file:`Containerfile` is
                written and where the build is performed. This value can be
                provided via the `tmp_path pytest fixture
                <https://docs.pytest.org/en/latest/how-to/tmp_path.html>`_
            rootdir_or_pytestconfig: root directory of the current test suite or
                a `pytestconfig fixture
                <https://docs.pytest.org/en/latest/reference/reference.html?highlight=pytestconfig#std-fixture-pytestconfig>`_
                object. This value is used to prepare the containers in
                :py:attr:`containers`.
            runtime: the container runtime to be used to perform the build. It
                can be retrieved using the
                :py:func:`pytest_container.plugin.container_runtime` fixture.
            target: an optional target to which the build will be run, see `the
                upstream documentation
                <https://docs.docker.com/develop/develop-images/multistage-build/#stop-at-a-specific-build-stage>`_
                for more information. Note that **no** verification of the
                :file:`Containerfile` is performed prior to the
                build. I.e. specifying an invalid target will fail your build.

        Returns:
            Id of the target container or of the last one (when no target was
            supplied) that was build
        """
        root = (
            rootdir_or_pytestconfig.rootpath
            if isinstance(rootdir_or_pytestconfig, Config)
            else rootdir_or_pytestconfig
        )
        self.prepare_build(
            tmp_path,
            runtime,
            root,
            extra_build_args,
        )
        return MultiStageBuild.run_build_step(
            tmp_path, runtime, target, extra_build_args
        )



# pylint: disable=missing-function-docstring,missing-module-docstring
from pathlib import Path
from tempfile import gettempdir
from typing import Optional
from typing import Union

import pytest

from pytest_container import Container
from pytest_container import DerivedContainer
from pytest_container.container import ContainerLauncher
from pytest_container.container import ImageFormat
from pytest_container.runtime import OciRuntimeBase

from . import images


def test_derived_container_fails_without_base() -> None:
    """Ensure that a DerivedContainer cannot be instantiated without providing
    the base parameter.

    """
    with pytest.raises(ValueError) as val_err_ctx:
        DerivedContainer()

    assert str(val_err_ctx.value) == "A base container must be provided"


def test_get_base_of_derived_container() -> None:
    """Ensure that :py:meth:`~pytest_container.DerivedContainer.get_base`
    returns a :py:class:`Container` with the correct url.

    """
    url = "registry.foobar.org/my_img:latest"
    assert DerivedContainer(base=url).get_base() == Container(url=url)


def test_image_format() -> None:
    """Check that the string representation of the ImageFormat enum is correct."""
    assert str(ImageFormat.DOCKER) == "docker"
    assert str(ImageFormat.OCIv1) == "oci"


def test_local_image_url(container_runtime: OciRuntimeBase) -> None:
    url = "docker.io/library/iDontExistHopefully/bazbarf/something"
    cont = Container(url=f"containers-storage:{url}")
    assert cont.local_image
    assert cont.url == url
    # prepare must not call `$runtime pull` as that would fail
    cont.prepare_container(container_runtime, Path("."), [])


def test_lockfile_path(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    """Check that the attribute
    :py:attr:`~pytest_container.ContainerBase.lockfile_filename` does change by
    the container having the attribute
    :py:attr:`~pytest_container.ContainerBase.container_id` set.

    """
    cont = DerivedContainer(
        base=images.OPENSUSE_BUSYBOX_URL, containerfile="ENV BAZ=1"
    )
    original_lock_fname = cont.filelock_filename

    cont.prepare_container(container_runtime, pytestconfig.rootpath)
    assert cont.container_id, "container_id must not be empty"
    assert cont.filelock_filename == original_lock_fname


def test_lockfile_unique() -> None:
    cont1 = DerivedContainer(
        base=images.OPENSUSE_BUSYBOX_URL, containerfile=""
    )
    cont2 = DerivedContainer(
        base=images.OPENSUSE_BUSYBOX_URL, containerfile="ENV foobar=1"
    )
    assert cont1.filelock_filename != cont2.filelock_filename


def test_removed_lockfile_does_not_kill_launcher(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    """Test that the container launcher doesn't die if the container lockfile
    got removed by another thread.

    It can happen in certain scenarios that a ``Container`` is launched from two
    concurrently running test functions. These test functions will use the same
    lockfile. A nasty data race can occur, where both test functions unlock the
    lockfile nearly at the same time, but then only one of them can succeed in
    removing it and the other test inadvertently fails. This is a regression
    test, that such a situation is tolerated and doesn't cause a failure.

    In this test we create a singleton container where we utilize that the
    lockfile is removed in ``__exit__()``. We hence already delete the lockfile
    in the ``with`` block and provoke a failure in ``__exit__()``.

    See also https://github.com/dcermak/pytest_container/issues/232.

    """
    cont = Container(url=images.LEAP_URL, singleton=True)

    with ContainerLauncher.from_pytestconfig(
        cont, container_runtime, pytestconfig
    ) as launcher:
        launcher.launch_container()

        lockfile_abspath = Path(gettempdir()) / cont.filelock_filename
        assert lockfile_abspath.exists

        lockfile_abspath.unlink()


def test_derived_container_build_tag(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    cont = DerivedContainer(base=images.OPENSUSE_BUSYBOX_URL)
    cont.prepare_container(container_runtime, pytestconfig.rootpath)
    assert cont._build_tag == images.OPENSUSE_BUSYBOX_URL


@pytest.mark.parametrize(
    "container_instance,url",
    [
        (Container(url=images.LEAP_URL), images.LEAP_URL),
        (images.LEAP_WITH_MAN, images.LEAP_URL),
        (images.LEAP_WITH_MAN_AND_LUA, images.LEAP_URL),
        (Container(url="containers-storage:foobar"), None),
        (
            DerivedContainer(base=Container(url="containers-storage:foobar")),
            None,
        ),
    ],
)
def test_baseurl(
    container_instance: Union[DerivedContainer, Container], url: Optional[str]
) -> None:
    assert container_instance.baseurl == url


def test_url_does_not_loose_containers_storage_part():
    local_prefix = "containers-storage"
    path = f"this/is/a/fake/image/with/{local_prefix}:latest"
    ctr = Container(url=f"{local_prefix}:{path}")
    assert ctr.local_image
    assert ctr.url == path



"""Module for managing podman pods."""

import contextlib
import json
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from subprocess import check_output
from types import TracebackType
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from _pytest.mark import ParameterSet
from pytest import Config

from pytest_container.container import Container
from pytest_container.container import ContainerData
from pytest_container.container import ContainerLauncher
from pytest_container.container import DerivedContainer
from pytest_container.container import create_host_port_port_forward
from pytest_container.container import lock_host_port_search
from pytest_container.helpers import get_extra_build_args
from pytest_container.helpers import get_extra_pod_create_args
from pytest_container.helpers import get_extra_run_args
from pytest_container.inspect import PortForwarding
from pytest_container.logging import _logger
from pytest_container.runtime import PodmanRuntime
from pytest_container.runtime import get_selected_runtime


@dataclass
class Pod:
    """A pod is a collection of containers that share the same network and port
    forwards. Currently only :command:`podman` supports creating pods.

    **Caution**: port forwards of the individual containers are ignored and only
    the port forwards in the pod class are taken into account!

    """

    #: containers belonging to the pod
    containers: List[Union[DerivedContainer, Container]]

    #: ports exposed by the pod
    forwarded_ports: List[PortForwarding] = field(default_factory=list)


@dataclass(frozen=True)
class PodData:
    """Class that is returned by the ``pod`` and ``pod_per_test`` fixtures. It
    contains all necessary information about the created pod and the containers
    running inside it.

    """

    #: The actual pod that has been launched
    pod: Pod

    #: The :py:class:`~pytest_container.container.ContainerData` instances of
    #: each container in the pod.
    container_data: List[ContainerData]

    #: unique id/hash of the running pod
    pod_id: str

    #: unique id/hash of the infra container of the pod
    infra_container_id: str

    #: ports exposed by this pod
    forwarded_ports: List[PortForwarding]


def infra_container_id_from_pod_inspect(inspect_output: bytes) -> str:
    """Given the output of :command:`podman pod inspect $id`, return the id of
    the infra container.

    """
    # we don't want to directly query the infra container id via
    # podman pod inspect -f "{{.InfraContainerID}}"
    # as that doesn't work on ancient podman versions
    #
    # But both new and old podman versions have the Containers field with
    # (at this stage), just the infra container.
    # So we just grab the id from the full inspect
    pod_inspect = json.loads(inspect_output.decode().strip())

    # for $reasons, since podman 5, the output of `podman pod inspect $id`
    # is no longer a dict, but a list of a dict 😡
    infra_container = (
        pod_inspect[0] if isinstance(pod_inspect, list) else pod_inspect
    )["Containers"][0]

    # old podman had the id of the containers with lowercase, new podman has
    # uppercase => have to check for both :-(
    return str(infra_container.get("Id", infra_container.get("id")))


@dataclass
class PodLauncher:
    """A context manager that creates, starts and destroys a pod with all of its
    containers.

    """

    #: the pod that should be created and launched including all of its
    #: containers
    pod: Pod

    #: root directory of the pytest testsuite
    rootdir: Path

    #: optional name of the pod
    pod_name: str = ""

    #: additional arguments to pass to the container build commands
    extra_build_args: List[str] = field(default_factory=list)

    #: additional arguments to pass to the container run commands
    extra_run_args: List[str] = field(default_factory=list)

    #: additional arguments to pass to the ``pod create`` command
    extra_pod_create_args: List[str] = field(default_factory=list)

    _launchers: List[ContainerLauncher] = field(default_factory=list)
    _new_port_forwards: List[PortForwarding] = field(default_factory=list)

    #: id of the pod
    _pod_id: Optional[str] = None

    #: id of the infra pod
    _infra_container_id: Optional[str] = None

    _stack: contextlib.ExitStack = field(default_factory=contextlib.ExitStack)

    @staticmethod
    def from_pytestconfig(
        pod: Pod, pytestconfig: Config, pod_name: str = ""
    ) -> "PodLauncher":
        """Constructor of :py:class:`PodLauncher` that obtains the attributes
        :py:attr:`rootdir`, :py:attr:`extra_build_args`,
        :py:attr:`extra_run_args` and :py:attr:`extra_pod_create_args` from the
        pytest configuration object.

        """
        return PodLauncher(
            pod,
            pytestconfig.rootpath,
            extra_build_args=get_extra_build_args(pytestconfig),
            extra_run_args=get_extra_run_args(pytestconfig),
            extra_pod_create_args=get_extra_pod_create_args(pytestconfig),
            pod_name=pod_name,
        )

    def __enter__(self) -> "PodLauncher":
        runtime = get_selected_runtime()
        if runtime != PodmanRuntime():
            raise RuntimeError(
                f"pods can only be created with podman, but got {runtime}"
            )
        return self

    def launch_pod(self) -> None:
        """Creates the actual pod, establishes the port bindings and launches
        all containers in the pod.

        """
        runtime = get_selected_runtime()
        create_cmd = (
            [runtime.runner_binary, "pod", "create"]
            + (["--name", self.pod_name] if self.pod_name else [])
            + self.extra_pod_create_args
        )

        if self.pod.forwarded_ports:
            with lock_host_port_search(self.rootdir):
                self._new_port_forwards = create_host_port_port_forward(
                    self.pod.forwarded_ports
                )
                # port_forward_args = []
                for new_forward in self._new_port_forwards:
                    create_cmd += new_forward.forward_cli_args

                _logger.debug("Creating pod via: %s", create_cmd)
                self._pod_id = check_output(create_cmd).decode().strip()
        else:
            _logger.debug("Creating pod via: %s", create_cmd)
            self._pod_id = check_output(create_cmd).decode().strip()

        def _delete_pod() -> None:
            if self._pod_id:
                _logger.debug("Removing pod %s", self._pod_id)
                check_output(
                    [runtime.runner_binary, "pod", "rm", "-f", self._pod_id]
                )
            else:
                _logger.debug("Not removing pod, not created")

        self._stack.callback(_delete_pod)

        self._infra_container_id = infra_container_id_from_pod_inspect(
            check_output(
                [
                    runtime.runner_binary,
                    "pod",
                    "inspect",
                    self._pod_id,
                ]
            )
        )

        for container in self.pod.containers:
            self._launchers.append(
                self._stack.enter_context(
                    ContainerLauncher(
                        container=container,
                        container_runtime=runtime,
                        rootdir=self.rootdir,
                        extra_build_args=self.extra_build_args,
                        extra_run_args=(
                            ["--pod", self._pod_id] + self.extra_run_args
                        ),
                        _expose_ports=False,
                    )
                )
            )
            self._launchers[-1].launch_container()

        assert len(self.pod.containers) == len(self._launchers)

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> None:
        self._stack.close()
        self._pod_id = None
        self._infra_container_id = None

    @property
    def pod_data(self) -> PodData:
        """Returns the :py:class:`PodData` corresponding to this podman pod."""
        if not self._pod_id or not self._infra_container_id:
            raise RuntimeError("Pod has not been created")

        return PodData(
            pod=self.pod,
            container_data=[
                launcher.container_data for launcher in self._launchers
            ],
            pod_id=self._pod_id,
            infra_container_id=self._infra_container_id,
            forwarded_ports=self._new_port_forwards,
        )


def pod_from_pytest_param(param: Union[ParameterSet, Pod]) -> Pod:
    """Extracts the :py:class:`~pytest_container.pod.Pod` from a `pytest.param
    <https://docs.pytest.org/en/stable/reference.html?#pytest.param>`_ or just
    returns the value directly, if it is a
    :py:class:`~pytest_container.pod.Pod`.

    """
    if isinstance(param, Pod):
        return param

    if len(param.values) > 0 and isinstance(param.values[0], Pod):
        return param.values[0]

    raise ValueError(f"Invalid pytest.param values: {param.values}")



# pylint: disable=missing-function-docstring,missing-module-docstring
from pathlib import Path
from typing import Union

import pytest
from pytest import Config

from pytest_container import Container
from pytest_container import DerivedContainer
from pytest_container import get_extra_build_args
from pytest_container.build import MultiStageBuild
from pytest_container.container import BindMount
from pytest_container.container import ContainerData
from pytest_container.container import ContainerLauncher
from pytest_container.container import EntrypointSelection
from pytest_container.inspect import PortForwarding
from pytest_container.runtime import LOCALHOST
from pytest_container.runtime import OciRuntimeBase

from .images import LEAP
from .images import LEAP_URL
from .images import LEAP_WITH_MAN
from .images import LEAP_WITH_MAN_AND_LUA
from .images import OPENSUSE_BUSYBOX_URL

TAG1 = "local/foobar/bazbarf"


LEAP_WITH_TAG = DerivedContainer(
    base=LEAP_URL,
    add_build_tags=[TAG1, "localhost/opensuse/leap/man:latest"],
)


BUSYBOX_WITH_ENTRYPOINT = Container(
    url=OPENSUSE_BUSYBOX_URL,
    custom_entry_point="/bin/sh",
)
#: This is just a busybox container with 4MB of random data in there
BUSYBOX_WITH_GARBAGE = DerivedContainer(
    base=BUSYBOX_WITH_ENTRYPOINT,
    containerfile="""RUN dd if=/dev/random of=/foobar bs=4M count=1
""",
)

SLEEP_CONTAINER = DerivedContainer(
    base=LEAP_URL,
    containerfile="""
# ps is needed for the tests to filter processes
RUN zypper -n in ps
ENTRYPOINT ["/usr/bin/sleep", "3600"]""",
)

LEAP2 = DerivedContainer(base=LEAP)

LEAP_WITH_BIN = DerivedContainer(
    base=LEAP,
    containerfile="""RUN zypper -n in system-user-bin
USER bin
""",
)

LEAP_WITH_BIN_AND_CMD = DerivedContainer(
    base=LEAP_WITH_BIN,
    containerfile="""CMD ["/usr/bin/sleep", "3600"]""",
    forwarded_ports=[PortForwarding(container_port=8080)],
)

CONTAINER_IMAGES = [LEAP, LEAP_WITH_MAN, LEAP_WITH_MAN_AND_LUA]

MULTI_STAGE_BUILD = MultiStageBuild(
    containers={
        "builder": LEAP_WITH_MAN,
        "runner1": LEAP,
        "runner2": "docker.io/alpine",
    },
    containerfile_template=r"""FROM $builder as builder
WORKDIR /src
RUN echo $$'#!/bin/sh \n\
echo "foobar"' > test.sh && chmod +x test.sh

FROM $runner1 as runner1
WORKDIR /bin
COPY --from=builder /src/test.sh .
ENTRYPOINT ["/bin/test.sh"]

FROM $runner2 as runner2
WORKDIR /bin
COPY --from=builder /src/test.sh .
""",
)

# This container would just stop if we would launch it with -d and use the
# default entrypoint. If we set the entrypoint to bash, then it should stay up.
CONTAINER_THAT_STOPS = DerivedContainer(
    base=LEAP,
    containerfile="""ENTRYPOINT ["/bin/echo", "hello world"]""",
    entry_point=EntrypointSelection.BASH,
)


@pytest.mark.parametrize("container", [LEAP], indirect=["container"])
def test_leap(container: ContainerData):
    assert container.connection.file("/etc/os-release").exists
    assert not container.connection.exists("man")
    assert not container.connection.exists("lua")


@pytest.mark.parametrize("container", [LEAP], indirect=["container"])
def test_container_data(container: ContainerData):
    assert container.container_id
    assert container.image_url_or_id == LEAP.url
    assert container.container == LEAP


def test_local_container_image_ref(
    container_runtime: OciRuntimeBase, pytestconfig: Config
):
    LEAP_WITH_TAG.prepare_container(container_runtime, pytestconfig.rootpath)

    # this container only works if LEAP_WITH_TAG exists already
    local_container = Container(url=f"containers-storage:{TAG1}")

    with ContainerLauncher(
        local_container, container_runtime, pytestconfig.rootpath
    ) as launcher:
        launcher.launch_container()
        connection = launcher.container_data.connection
        assert connection.file("/etc/os-release").exists
        assert (
            'ID="opensuse-leap"'
            in connection.file("/etc/os-release").content_string
        )


@pytest.mark.parametrize("container", [LEAP_WITH_MAN], indirect=["container"])
def test_leap_with_man(container: ContainerData):
    assert container.connection.exists("man")
    assert not container.connection.exists("lua")


@pytest.mark.parametrize("container", [LEAP_WITH_MAN], indirect=["container"])
def test_derived_container_data(container: ContainerData):
    assert container.container_id
    assert container.image_url_or_id == LEAP_WITH_MAN.container_id
    assert container.container == LEAP_WITH_MAN


@pytest.mark.parametrize("container", [LEAP2], indirect=True)
def test_container_without_containerfile_and_without_tags_not_rebuild(
    container: ContainerData,
):
    assert (
        isinstance(container.container, DerivedContainer)
        and not container.container.containerfile
        and not container.container.add_build_tags
    )
    assert container.container.get_base() == LEAP
    assert container.container.url == LEAP.url
    assert container.container._build_tag == LEAP.url


@pytest.mark.parametrize("container", [LEAP_WITH_TAG], indirect=True)
def test_container_without_containerfile_but_with_tags_is_rebuild(
    container: ContainerData,
):
    assert (
        isinstance(container.container, DerivedContainer)
        and not container.container.containerfile
        and container.container.add_build_tags
    )
    assert container.container.get_base() == LEAP
    assert container.container.container_id != LEAP.url


@pytest.mark.parametrize(
    "container", [LEAP_WITH_MAN_AND_LUA], indirect=["container"]
)
def test_leap_with_man_and_lua(container: ContainerData):
    assert container.connection.exists("man")
    assert container.connection.exists("lua")


@pytest.mark.parametrize(
    "cont,base", list(zip(CONTAINER_IMAGES, [LEAP, LEAP, LEAP_WITH_MAN]))
)
def test_container_objects(
    cont: Union[Container, DerivedContainer],
    base: Union[Container, DerivedContainer],
) -> None:
    assert cont.get_base() == base


def test_auto_container_fixture(auto_container: ContainerData):
    assert auto_container.connection.file("/etc/os-release").exists


@pytest.mark.parametrize(
    "container", [BUSYBOX_WITH_ENTRYPOINT], indirect=["container"]
)
def test_custom_entry_point(container: ContainerData):
    container.connection.check_output("true")


@pytest.mark.parametrize(
    "container", [SLEEP_CONTAINER], indirect=["container"]
)
def test_default_entry_point(container: ContainerData):
    sleep = container.connection.process.filter(comm="sleep")
    assert len(sleep) == 1
    assert "/usr/bin/sleep 3600" == sleep[0].args


@pytest.mark.parametrize("container", [CONTAINER_THAT_STOPS], indirect=True)
def test_container_that_stops(container: ContainerData) -> None:
    # it should just be alive
    container.connection.check_output("true")


def test_container_size(
    container_runtime: OciRuntimeBase, pytestconfig: Config
):
    for container in [BUSYBOX_WITH_ENTRYPOINT, BUSYBOX_WITH_GARBAGE]:
        container.prepare_container(container_runtime, pytestconfig.rootpath)

    assert container_runtime.get_image_size(
        BUSYBOX_WITH_ENTRYPOINT
    ) < container_runtime.get_image_size(BUSYBOX_WITH_GARBAGE)
    assert (
        container_runtime.get_image_size(BUSYBOX_WITH_ENTRYPOINT)
        - container_runtime.get_image_size(BUSYBOX_WITH_GARBAGE)
        < 4096 * 1024 * 1024
    )


@pytest.mark.parametrize(
    "container",
    (LEAP_WITH_BIN, LEAP_WITH_BIN_AND_CMD),
    indirect=True,
)
def test_derived_containers_use_correct_user(
    container: ContainerData,
) -> None:
    assert container.connection.check_output("id -nu").strip() == "bin"
    assert container.inspect.config.user == "bin"


@pytest.mark.parametrize(
    "container",
    [
        DerivedContainer(base=ctr, extra_launch_args=["--user", "root"])
        for ctr in (LEAP_WITH_BIN, LEAP_WITH_BIN_AND_CMD)
    ],
    indirect=True,
)
def test_derived_container_respects_launch_args(
    container: ContainerData,
) -> None:
    assert int(container.connection.check_output("id -u").strip()) == 0


def test_multistage_containerfile() -> None:
    assert "FROM docker.io/alpine" in MULTI_STAGE_BUILD.containerfile


def test_multistage_build(
    tmp_path: Path, pytestconfig: Config, container_runtime: OciRuntimeBase
):
    MULTI_STAGE_BUILD.build(
        tmp_path,
        pytestconfig.rootpath,
        container_runtime,
        extra_build_args=get_extra_build_args(pytestconfig),
    )


def test_multistage_build_target(
    tmp_path: Path, pytestconfig: Config, container_runtime: OciRuntimeBase
):
    first_target = MULTI_STAGE_BUILD.build(
        tmp_path,
        pytestconfig.rootpath,
        container_runtime,
        "runner1",
        extra_build_args=get_extra_build_args(pytestconfig),
    )
    assert (
        LOCALHOST.check_output(
            f"{container_runtime.runner_binary} run --rm {first_target}",
        ).strip()
        == "foobar"
    )

    second_target = MULTI_STAGE_BUILD.build(
        tmp_path,
        pytestconfig,
        container_runtime,
        "runner2",
        extra_build_args=get_extra_build_args(pytestconfig),
    )

    assert first_target != second_target
    assert (
        LOCALHOST.check_output(
            f"{container_runtime.runner_binary} run --rm {second_target} /bin/test.sh",
        ).strip()
        == "foobar"
    )

    for distro, target in (
        ("Leap", first_target),
        ("Alpine", second_target),
    ):
        assert (
            distro
            in LOCALHOST.check_output(
                f"{container_runtime.runner_binary} run --rm --entrypoint= {target} "
                "cat /etc/os-release",
            ).strip()
        )


LEAP_THAT_ECHOES_STUFF = DerivedContainer(
    base=LEAP, containerfile="""CMD ["echo", "foobar"]"""
)


@pytest.mark.parametrize("container", [LEAP_THAT_ECHOES_STUFF], indirect=True)
def test_container_logs(container: ContainerData) -> None:
    assert "foobar" in container.read_container_logs()


def test_container_stops_on_exit(
    container_runtime: OciRuntimeBase,
    pytestconfig: Config,
    tmp_path: Path,
):
    _CONTAINER_THAT_TRAPS_SIGTERM = DerivedContainer(
        base=BUSYBOX_WITH_ENTRYPOINT,
        containerfile="""COPY tests/files/entrypoint.sh /usr/local/bin/entrypoint.sh
RUN chmod +x /usr/local/bin/entrypoint.sh
CMD ["/usr/local/bin/entrypoint.sh"]
""",
        volume_mounts=[
            BindMount(host_path=tmp_path, container_path="/var/volume")
        ],
    )
    with ContainerLauncher(
        _CONTAINER_THAT_TRAPS_SIGTERM,
        container_runtime,
        pytestconfig.rootpath,
    ) as launcher:
        launcher.launch_container()
        connection = launcher.container_data.connection
        assert connection.file("/var/volume/cleanup_confirmed").exists

    assert (tmp_path / "cleanup_confirmed").read_text().strip() == "1"



"""Unit tests of the Version class"""

# pylint: disable=missing-function-docstring,missing-module-docstring
import pytest

from pytest_container import Version
from pytest_container.runtime import OciRuntimeBase
from pytest_container.runtime import _get_docker_version
from pytest_container.runtime import _get_podman_version

# pragma pylint: disable=missing-function-docstring


@pytest.mark.parametrize(
    "ver1,ver2",
    [
        (Version(1, 0, 2), Version(1, 0, 2)),
        (Version(2, 0), Version(2, 0, 0)),
    ],
)
def test_version_eq(ver1: Version, ver2: Version):
    assert ver1 == ver2


def test_incompatible_types_eq() -> None:
    assert Version(1, 2) != 3


def test_incompatible_types_cmp() -> None:
    with pytest.raises(TypeError) as ctx:
        _ = Version(1, 2) < 3

    assert "'<' not supported between instances of 'Version' and 'int'" in str(
        ctx.value
    )


@pytest.mark.parametrize(
    "ver1,ver2",
    [
        (Version(1, 0, 2), Version(1, 0, 1)),
        (Version(2, 0, 1), Version(1, 0, 1)),
        (Version(1, 5, 1), Version(1, 0, 1)),
        (Version(1, 0, 1), Version(1, 0, 1, build="foobar")),
        (Version(1, 0, 1), Version(1, 0, 1, release="foobar")),
        (Version(1, 0, 1), Version(1, 0, 1, release="foobar", build="5")),
    ],
)
def test_version_ne(ver1: Version, ver2: Version):
    assert ver1 != ver2


@pytest.mark.parametrize(
    "ver,stringified",
    [
        (Version(1, 2), "1.2"),
        (Version(1, 2, 5), "1.2.5"),
        (Version(0, 3, 0), "0.3.0"),
        (Version(1, 2, 5, build="sdf"), "1.2.5 build sdf"),
        (Version(1, 2, 5, release="1.fc16"), "1.2.5-1.fc16"),
        (
            Version(1, 2, 5, release="1.fc16", build="38"),
            "1.2.5-1.fc16 build 38",
        ),
    ],
)
def test_version_str(ver: Version, stringified: str):
    assert str(ver) == stringified


@pytest.mark.parametrize(
    "larger,smaller",
    [
        (Version(1, 2, 3), Version(1, 2, 2)),
        (Version(1, 2, 3), Version(1, 1, 3)),
        (Version(1, 2, 3), Version(0, 2, 3)),
    ],
)
def test_version_ge_gt(larger: Version, smaller: Version):
    assert larger > smaller
    assert not smaller > larger

    assert larger >= smaller
    assert not smaller >= larger

    # pragma pylint: disable=comparison-with-itself
    assert larger >= larger
    assert smaller >= smaller


@pytest.mark.parametrize(
    "larger,smaller",
    [
        (Version(1, 2, 3), Version(1, 2, 2)),
        (Version(1, 2, 3), Version(1, 1, 3)),
        (Version(1, 2, 3), Version(0, 2, 3)),
    ],
)
def test_version_le_lt(larger: Version, smaller: Version):
    assert smaller < larger
    assert not larger < smaller

    assert smaller <= larger
    assert not larger <= smaller

    # pragma pylint: disable=comparison-with-itself
    assert larger <= larger
    assert smaller <= smaller


@pytest.mark.parametrize(
    "stdout,ver",
    [
        (
            "Docker version 1.12.6, build 78d1802",
            Version(1, 12, 6, build="78d1802"),
        ),
        (
            "Docker version 20.10.12-ce, build 459d0dfbbb51",
            Version(20, 10, 12, build="459d0dfbbb51", release="ce"),
        ),
        (
            "Docker version 20.10.16, build aa7e414",
            Version(20, 10, 16, build="aa7e414"),
        ),
        (
            "Docker version 1.13.1, build 7d71120/1.13.1",
            Version(1, 13, 1, build="7d71120/1.13.1"),
        ),
        (
            "Docker version 20.10.17+azure-1, build 100c70180fde3601def79a59cc3e996aa553c9b9",
            Version(
                20,
                10,
                17,
                release="azure-1",
                build="100c70180fde3601def79a59cc3e996aa553c9b9",
            ),
        ),
    ],
)
def test_docker_version_extract(stdout: str, ver: Version):
    assert _get_docker_version(stdout) == ver


@pytest.mark.parametrize(
    "stdout,ver",
    [
        ("podman version 3.0.1", Version(3, 0, 1)),
        ("podman version 3.4.4", Version(3, 4, 4)),
        ("podman version 1.6.4", Version(1, 6, 4)),
        ("podman version 4.0.2", Version(4, 0, 2)),
    ],
)
def test_podman_version_extract(stdout: str, ver: Version):
    assert _get_podman_version(stdout) == ver


def test_container_runtime_parsing(host, container_runtime: OciRuntimeBase):
    """Test that we can recreate the output of
    :command:`$container_runtime_binary --version` from the attribute
    :py:attr:`~pytest_container.runtime.OciRuntimeBase.version`.

    """
    version_without_build = Version(
        major=container_runtime.version.major,
        minor=container_runtime.version.minor,
        patch=container_runtime.version.patch,
    )
    version_string = (
        host.run_expect([0], f"{container_runtime.runner_binary} --version")
        .stdout.strip()
        .lower()
    )

    assert (
        f"{container_runtime.runner_binary} version {version_without_build}"
        in version_string
    )

    if container_runtime.runner_binary == "docker":
        assert f"build {container_runtime.version.build}" in version_string


@pytest.mark.parametrize(
    "ver_str,expected_version",
    [
        ("1.2.3", Version(1, 2, 3)),
        ("65.8", Version(65, 8)),
        ("6", Version(6, 0)),
        ("2.7.8-55ubuntu~", Version(2, 7, 8, release="55ubuntu~")),
        ("2.8-16.fc37", Version(2, 8, release="16.fc37")),
        (
            "2.7.8-55ubuntu~ build 42",
            Version(2, 7, 8, release="55ubuntu~", build="42"),
        ),
        ("2.7.8 build 42", Version(2, 7, 8, build="42")),
    ],
)
def test_parse_valid_version_strings(ver_str: str, expected_version: Version):
    """Check that ``ver_str`` is parsed correctly into ``expected_version``."""
    assert Version.parse(ver_str) == expected_version


_INVALID_1 = "2.5 not-build 46"
_INVALID_2 = "16.84 build but with too much text"
_INVALID_3 = "asdf"


@pytest.mark.parametrize(
    "ver_str,exception,exception_text",
    [
        (
            _INVALID_1,
            ValueError,
            f"Invalid version string: {_INVALID_1}",
        ),
        (
            _INVALID_2,
            ValueError,
            f"Invalid version string: {_INVALID_2}",
        ),
        (_INVALID_3, ValueError, f"Invalid version string: {_INVALID_3}"),
    ],
)
def test_parse_invalid_version_strings(
    ver_str: str, exception, exception_text: str
):
    """Check that parsing the version string `ver_str` raises the supplied
    exception type and includes the provided ``exception_text``.

    """
    with pytest.raises(exception) as exc_ctx:
        Version.parse(ver_str)
    assert exception_text in str(exc_ctx.value)



"""The plugin module contains all fixtures that are provided by
``pytest_container``.

"""

import sys
from subprocess import PIPE
from subprocess import run
from typing import Callable
from typing import Generator

from pytest_container.container import ContainerData
from pytest_container.container import ContainerLauncher
from pytest_container.container import container_and_marks_from_pytest_param
from pytest_container.logging import _logger
from pytest_container.pod import PodData
from pytest_container.pod import PodLauncher
from pytest_container.pod import pod_from_pytest_param
from pytest_container.runtime import OciRuntimeBase
from pytest_container.runtime import get_selected_runtime

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from _pytest.config import Config
from _pytest.fixtures import SubRequest
from pytest import fixture
from pytest import skip


@fixture(scope="session")
def container_runtime() -> OciRuntimeBase:
    """pytest fixture that returns the currently selected container runtime
    according to the rules outlined :ref:`here <runtime selection rules>`.

    """
    return get_selected_runtime()


def _log_container_logs(
    container_id: str, ctr_runtime: OciRuntimeBase
) -> None:
    # don't die if logging fails for some reason
    # pylint: disable=subprocess-run-check
    logs_call = run(
        [ctr_runtime.runner_binary, "logs", container_id],
        stdout=PIPE,
        stderr=PIPE,
    )
    if logs_call.returncode == 0:
        _logger.debug(
            "logs from container %s: %s",
            container_id,
            logs_call.stdout.decode(),
        )


def _create_auto_container_fixture(
    scope: Literal["session", "function"],
) -> Callable[
    [SubRequest, OciRuntimeBase, Config], Generator[ContainerData, None, None]
]:
    def fixture_funct(
        request: SubRequest,
        # we must call this parameter container runtime, so that pytest will
        # treat it as a fixture, but that causes pylint to complain…
        # pylint: disable=redefined-outer-name
        container_runtime: OciRuntimeBase,
        pytestconfig: Config,
    ) -> Generator[ContainerData, None, None]:
        """Fixture that will build & launch a container that is either passed as a
        request parameter or it will be automatically parametrized via
        pytest_generate_tests.
        """

        try:
            container, _ = container_and_marks_from_pytest_param(request.param)
        except AttributeError as attr_err:
            raise RuntimeError(
                "This fixture was not parametrized correctly, "
                "did you forget to call `auto_container_parametrize` in `pytest_generate_tests`?"
            ) from attr_err
        _logger.debug("Requesting the container %s", str(container))

        if scope == "session" and container.singleton:
            raise RuntimeError(
                f"A singleton container ({container}) cannot be used in a session level fixture"
            )

        with ContainerLauncher.from_pytestconfig(
            container=container,
            container_runtime=container_runtime,
            pytestconfig=pytestconfig,
        ) as launcher:
            # we want to ensure that the container's logs are saved at "all
            # cost", especially when the container fails to launch for some
            # reason
            try:
                launcher.launch_container()
                container_data = launcher.container_data
                yield container_data
            finally:
                if launcher._container_id:
                    _log_container_logs(
                        launcher._container_id, container_runtime
                    )

    return fixture(scope=scope)(fixture_funct)


def _create_auto_pod_fixture(
    scope: Literal["session", "function"],
) -> Callable[
    [SubRequest, OciRuntimeBase, Config], Generator[PodData, None, None]
]:
    def fixture_funct(
        request: SubRequest,
        # we must call this parameter container runtime, so that pytest will
        # treat it as a fixture, but that causes pylint to complain…
        # pylint: disable=redefined-outer-name
        container_runtime: OciRuntimeBase,
        pytestconfig: Config,
    ) -> Generator[PodData, None, None]:
        if "podman" not in container_runtime.runner_binary:
            skip("Pods are only supported in podman")

        pod = pod_from_pytest_param(request.param)
        with PodLauncher.from_pytestconfig(pod, pytestconfig) as launcher:
            try:
                launcher.launch_pod()
                pod_data = launcher.pod_data
                yield pod_data
            finally:
                for ctr_launcher in launcher._launchers:
                    if ctr_launcher._container_id:
                        _log_container_logs(
                            ctr_launcher._container_id, container_runtime
                        )

    return fixture(scope=scope)(fixture_funct)


#: This fixture parametrizes the test function once for each container image
#: defined in the module level variable ``CONTAINER_IMAGES`` of the current test
#: module and yield an instance of
#: :py:attr:`~pytest_container.container.ContainerData`.
#: This fixture will reuse the same container for all tests of the same session.
auto_container = _create_auto_container_fixture("session")

#: Fixture that expects to be parametrized with an instance of a subclass of
#: :py:class:`~pytest_container.container.ContainerBase` with `indirect=True`.
#: It will launch the container and yield an instance of
#: :py:attr:`~pytest_container.container.ContainerData`.
#: This fixture will reuse the same container for all tests of the same session.
container = _create_auto_container_fixture("session")

#: Same as :py:func:`auto_container` but it will launch individual containers
#: for each test function.
auto_container_per_test = _create_auto_container_fixture("function")

#: Same as :py:func:`container` but it will launch individual containers for
#: each test function.
container_per_test = _create_auto_container_fixture("function")

#: Fixture that has to be parametrized with an instance of
#: :py:class:`~pytest_container.pod.Pod` with `indirect=True`.
#: It creates the pod, launches all of its containers and yields an instance of
#: :py:class:`~pytest_container.pod.PodData`. The fixture automatically skips
#: the test when the current container runtime is not :command:`podman`.
#: The pod created by this fixture is shared by all test functions.
pod = _create_auto_pod_fixture("session")

#: Same as :py:func:`pod`, except that it creates a pod for each test function.
pod_per_test = _create_auto_pod_fixture("function")



# pylint: disable=missing-function-docstring,missing-module-docstring
import os
import re
import subprocess
import tempfile
from contextlib import ExitStack
from pathlib import Path
from time import sleep
from typing import Any
from unittest.mock import patch

import pytest

from pytest_container import inspect
from pytest_container.container import BindMount
from pytest_container.container import Container
from pytest_container.container import ContainerData
from pytest_container.container import ContainerLauncher
from pytest_container.container import ContainerVolume
from pytest_container.container import DerivedContainer
from pytest_container.container import EntrypointSelection
from pytest_container.runtime import LOCALHOST
from pytest_container.runtime import OciRuntimeBase

from .images import CMDLINE_APP_CONTAINER
from .images import CONTAINER_THAT_FAILS_TO_LAUNCH
from .images import LEAP
from .test_volumes import LEAP_WITH_BIND_MOUNT_AND_VOLUME
from .test_volumes import LEAP_WITH_CONTAINER_VOLUMES
from .test_volumes import LEAP_WITH_VOLUMES

LEAP_WITH_STOPSIGNAL_SIGKILL = DerivedContainer(
    base=LEAP,
    containerfile="STOPSIGNAL SIGKILL",
    entry_point=EntrypointSelection.BASH,
)

LEAP_WITH_STOPSIGNAL_SIGKILL_AND_ENTRYPOINT = DerivedContainer(
    base=LEAP,
    containerfile="STOPSIGNAL SIGKILL",
    entry_point=EntrypointSelection.IMAGE,
)

LEAP_WITH_STOPSIGNAL_SIGKILL_AND_CUSTOM_ENTRYPOINT = DerivedContainer(
    base=LEAP,
    containerfile="STOPSIGNAL SIGKILL",
    custom_entry_point="/bin/sh",
)

PYTHON_LEAP = DerivedContainer(
    base=LEAP,
    containerfile="""
RUN set -euxo pipefail; zypper -n ref; zypper -n in python3 curl;

ENTRYPOINT ["/usr/bin/python3"]
CMD ["-m", "http.server"]
""",
)


def _test_func(con: Any) -> None:
    sleep(5)
    assert "Leap" in con.run_expect([0], "cat /etc/os-release").stdout


@pytest.mark.parametrize("container", [LEAP], indirect=True)
def test_cleanup_not_immediate(container: ContainerData) -> None:
    _test_func(container.connection)


@pytest.mark.parametrize("container_per_test", [LEAP], indirect=True)
def test_cleanup_not_immediate_per_test(
    container_per_test: ContainerData,
) -> None:
    _test_func(container_per_test.connection)


@pytest.mark.parametrize(
    "cont",
    [
        LEAP_WITH_VOLUMES,
        LEAP_WITH_CONTAINER_VOLUMES,
        LEAP_WITH_BIND_MOUNT_AND_VOLUME,
    ],
)
def test_launcher_creates_and_cleanes_up_volumes(
    cont: DerivedContainer,
    pytestconfig: pytest.Config,
    container_runtime: OciRuntimeBase,
) -> None:
    with ContainerLauncher.from_pytestconfig(
        cont, container_runtime, pytestconfig
    ) as launcher:
        launcher.launch_container()

        container = launcher.container_data.container
        assert container.volume_mounts

        for vol in container.volume_mounts:
            if isinstance(vol, BindMount):
                assert vol.host_path and os.path.exists(vol.host_path)
            elif isinstance(vol, ContainerVolume):
                assert vol.volume_id
                assert LOCALHOST.run_expect(
                    [0],
                    f"{container_runtime.runner_binary} volume inspect {vol.volume_id}",
                )
            else:
                assert False, f"invalid volume type {type(vol)}"

    for vol in container.volume_mounts:
        if isinstance(vol, BindMount):
            assert not vol.host_path
        elif isinstance(vol, ContainerVolume):
            assert not vol.volume_id
        else:
            assert False, f"invalid volume type {type(vol)}"


LEAP_WITH_VOLUME_IN_DOCKERFILE = DerivedContainer(
    base=LEAP, containerfile="VOLUME /foo"
)


@pytest.mark.parametrize("cont", [LEAP_WITH_VOLUME_IN_DOCKERFILE])
def test_launcher_cleanes_up_volumes_from_image(
    cont: DerivedContainer,
    pytestconfig: pytest.Config,
    container_runtime: OciRuntimeBase,
    host: Any,
) -> None:
    with ContainerLauncher.from_pytestconfig(
        cont, container_runtime, pytestconfig
    ) as launcher:
        launcher.launch_container()

        container = launcher.container_data.container
        assert not container.volume_mounts

        mounts = launcher.container_data.inspect.mounts
        assert (
            len(mounts) == 1
            and isinstance(mounts[0], inspect.VolumeMount)
            and mounts[0].destination == "/foo"
        )

        vol_name = mounts[0].name
    assert (
        "no such volume"
        in host.run_expect(
            [1, 125],
            f"{container_runtime.runner_binary} volume inspect {vol_name}",
        ).stderr.lower()
    )


def test_launcher_container_data_not_available_after_exit(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    with ContainerLauncher.from_pytestconfig(
        LEAP, container_runtime, pytestconfig
    ) as launcher:
        launcher.launch_container()
        assert launcher.container_data

    with pytest.raises(RuntimeError) as runtime_err_ctx:
        _ = launcher.container_data

    assert f"{LEAP} has not started" in str(runtime_err_ctx.value)


def test_launcher_fails_on_failing_healthcheck(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config, host
):
    container_name = "container_with_failing_healthcheck"
    with pytest.raises(RuntimeError) as runtime_err_ctx:
        with ContainerLauncher.from_pytestconfig(
            container=CONTAINER_THAT_FAILS_TO_LAUNCH,
            container_runtime=container_runtime,
            pytestconfig=pytestconfig,
            container_name=container_name,
        ) as launcher:
            launcher.launch_container()
            assert False, "This code must be unreachable"

    err_msg_regex = re.compile(
        r"Container (\d|\w*) did not become healthy within (\d+\.\d+s),"
        r" took (\d+.\d+s) and state is (\w+)"
    )
    err_msg_match = err_msg_regex.match(str(runtime_err_ctx.value))
    assert err_msg_match, (
        f"Error message '{str(runtime_err_ctx.value)}' does "
        + "not match expected pattern {err_msg_regex}"
    )

    # the container must not exist anymore
    err_msg = host.run_expect(
        [1, 125],
        f"{container_runtime.runner_binary} inspect {container_name}",
    ).stderr
    assert ("no such object" in err_msg.lower()) or (
        "error getting image" in err_msg
    )


@pytest.mark.parametrize(
    "container", [LEAP_WITH_STOPSIGNAL_SIGKILL], indirect=True
)
def test_launcher_overrides_stopsignal(container: ContainerData) -> None:
    """Verify that we override the stop signal by default to ``SIGTERM`` as we
    launch containers with :file:`/bin/bash` as the entrypoint.

    """
    assert container.inspect.config.stop_signal in (15, "SIGTERM")


@pytest.mark.parametrize(
    "container",
    [
        LEAP_WITH_STOPSIGNAL_SIGKILL_AND_ENTRYPOINT,
        LEAP_WITH_STOPSIGNAL_SIGKILL_AND_CUSTOM_ENTRYPOINT,
    ],
    indirect=True,
)
def test_launcher_does_not_override_stopsignal_for_entrypoint(
    container: ContainerData,
) -> None:
    """Check that the stop signal is **not** modified when the attribute
    `default_entry_point` is ``True`` (then we assume that the stop signal has
    been set to the appropriate value by the author of the image).

    """
    assert container.inspect.config.stop_signal in (9, "SIGKILL")


@pytest.mark.parametrize("container", [CMDLINE_APP_CONTAINER], indirect=True)
def test_launcher_does_can_check_binaries_with_entrypoint(
    container: ContainerData,
) -> None:
    """Check that the we can check for installed binaries even if the container
    has an entrypoint specified that is not a shell and terminates immediately.
    """
    assert container.connection.exists("bash")


def test_derived_container_pulls_base(
    container_runtime: OciRuntimeBase, host: Any, pytestconfig: pytest.Config
) -> None:
    registry_url = "registry.opensuse.org/opensuse/registry:latest"

    # remove the container image so that the preparation in the launcher must
    # pull the image
    host.run(f"{container_runtime.runner_binary} rmi {registry_url}")

    reg = DerivedContainer(base=registry_url)
    with ContainerLauncher.from_pytestconfig(
        reg, container_runtime, pytestconfig
    ) as launcher:
        launcher.launch_container()
        assert launcher.container_data.container_id


def test_pulls_container(
    container_runtime: OciRuntimeBase,
    pytestconfig: pytest.Config,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test of the pull-behavior switching via the environment variable
    ``PULL_ALWAYS``

    """
    quay_busybox = "quay.io/libpod/busybox"

    with ExitStack() as stack:
        # mock setup
        mock_check_output = stack.enter_context(
            patch("pytest_container.container.check_output")
        )
        mock_check_call = stack.enter_context(
            patch("pytest_container.container.call")
        )
        mock_check_output.return_value = None

        def _pull():
            Container(url=quay_busybox).prepare_container(
                container_runtime, pytestconfig.rootpath
            )

        # first test: should always pull the image
        monkeypatch.setenv("PULL_ALWAYS", "1")
        _pull()

        mock_check_output.assert_called_once_with(
            [container_runtime.runner_binary, "pull", quay_busybox]
        )
        mock_check_call.assert_not_called()

        mock_check_output.reset_mock()
        mock_check_call.reset_mock()

        # second test: should only pull the image if inspect fails
        # in this case we mock the inspect call to return 0, i.e. image is there
        monkeypatch.setenv("PULL_ALWAYS", "0")
        mock_check_call.return_value = 0

        _pull()
        mock_check_call.assert_called_once_with(
            [container_runtime.runner_binary, "inspect", quay_busybox]
        )
        mock_check_output.assert_not_called()

        mock_check_output.reset_mock()
        mock_check_call.reset_mock()

        # third test: pull the image if inspect fails, so we mock the inspect
        # call to return 1
        mock_check_call.return_value = 1

        _pull()
        mock_check_call.assert_called_once_with(
            [container_runtime.runner_binary, "inspect", quay_busybox]
        )
        mock_check_output.assert_called_once_with(
            [container_runtime.runner_binary, "pull", quay_busybox]
        )


def test_launcher_unlocks_on_preparation_failure(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    container_with_wrong_url = Container(
        url="registry.invalid.xyz.foobar/i/should/not/exist:42"
    )

    def try_launch():
        with pytest.raises(subprocess.CalledProcessError):
            with ContainerLauncher.from_pytestconfig(
                container_with_wrong_url,
                container_runtime,
                pytestconfig,
            ) as launcher:
                launcher.launch_container()
                assert False, "The container must not have launched"

    try_launch()
    # not the best as we are testing an internal implementation detail
    assert not Path(
        tempfile.gettempdir(), container_with_wrong_url.filelock_filename
    ).exists()


@pytest.mark.parametrize(
    "container,port_num",
    [
        (
            DerivedContainer(
                base=PYTHON_LEAP,
                extra_entrypoint_args=["-m", "http.server", "8080"],
            ),
            8080,
        ),
        (PYTHON_LEAP, 8000),
    ],
    indirect=["container"],
)
def test_extra_command_args(container: ContainerData, port_num: int) -> None:
    assert container.connection.check_output(
        f"curl -sf --retry 5 --retry-connrefused http://localhost:{port_num}"
    )



"""This module contains the class definitions that represent the output of
:command:`$runtime inspect $ctr_id`.

"""

import enum
import socket
from dataclasses import dataclass
from dataclasses import field
from datetime import timedelta
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict


@enum.unique
class NetworkProtocol(enum.Enum):
    """Network protocols supporting port forwarding."""

    #: Transmission Control Protocol
    TCP = "tcp"
    #: User Datagram Protocol
    UDP = "udp"

    def __str__(self) -> str:
        return self.value

    @property
    def SOCK_CONST(self) -> int:
        """Returns the appropriate socket type constant (``SOCK_STREAM`` or
        ``SOCK_DGRAM``) for the current protocol.

        """
        return {
            NetworkProtocol.TCP.value: socket.SOCK_STREAM,
            NetworkProtocol.UDP.value: socket.SOCK_DGRAM,
        }[self.value]


@dataclass(frozen=True)
class PortForwarding:
    """Representation of a port forward from a container to the host.

    To expose a port of a container automatically, create an instance of this
    class, set the attribute :py:attr:`container_port` and optionally
    :py:attr:`protocol` as well and pass it via the parameter
    :py:attr:`~pytest_container.container.ContainerBase.forwarded_ports` to
    either the :py:class:`pytest_container.container.Container` or
    :py:class:`pytest_container.container.DerivedContainer`:

    >>> Container(url="my-webserver", forwarded_ports=[PortForwarding(container_port=8000)])

    """

    #: The port which shall be exposed by the container.
    container_port: int

    #: The protocol which the exposed port is using. Defaults to TCP.
    protocol: NetworkProtocol = NetworkProtocol.TCP

    #: The port as which the port from :py:attr:`container_port` is exposed on
    #: the host. This value is automatically set by the `*container_*` fixtures,
    #: so there's no need for the user to modify it
    host_port: int = -1

    #: The IP address to which to bind. By default, it will be '::' (all addresses).
    bind_ip: str = ""

    @property
    def forward_cli_args(self) -> List[str]:
        """Returns a list of command line arguments for the container launch
        command to automatically expose this port forwarding.

        """

        if self.bind_ip:
            # If it contains a colon, it must be an IPv6 address and thus must
            # be wrapped in brackets for the launch command
            if ":" in self.bind_ip:
                bind_ip = f"[{self.bind_ip}]:"
            else:
                bind_ip = self.bind_ip + ":"
        else:
            bind_ip = ""

        return [
            "-p",
            bind_ip
            + ("" if self.host_port == -1 else f"{self.host_port}:")
            + f"{self.container_port}/{self.protocol}",
        ]

    def __str__(self) -> str:
        return str(self.forward_cli_args)


class ContainerInspectHealthCheck(TypedDict, total=False):
    """Dictionary created by loading the json output of :command:`podman inspect
    $img_id | jq '.[0]["Healthcheck]` or :command:`docker inspect $img_id | jq
    '.[0]["Config"]["Healthcheck]`.

    """

    Test: List[str]
    Interval: int
    Timeout: int
    StartPeriod: int
    Retries: int


@enum.unique
class ContainerHealth(enum.Enum):
    """Possible states of a container's health using the `HEALTHCHECK
    <https://docs.docker.com/engine/reference/builder/#healthcheck>`_ property
    of a container image.

    """

    #: the container has no health check defined
    NO_HEALTH_CHECK = ""
    #: the container is healthy
    HEALTHY = "healthy"
    #: the health check did not complete yet or did not fail often enough
    STARTING = "starting"
    #: the healthcheck failed
    UNHEALTHY = "unhealthy"


_DEFAULT_START_PERIOD = timedelta(seconds=0)
_DEFAULT_INTERVAL = timedelta(seconds=30)
_DEFAULT_TIMEOUT = timedelta(seconds=30)
_DEFAULT_RETRIES = 3


@dataclass(frozen=True)
class HealthCheck:
    """The HEALTHCHECK of a container image."""

    #: startup period of the container during which healthcheck failures will
    #: not count towards the failure count
    start_period: timedelta = field(default=_DEFAULT_START_PERIOD)

    #: healthcheck command is run every interval
    interval: timedelta = field(default=_DEFAULT_INTERVAL)

    #: timeout of the healthcheck command after which it is considered unsuccessful
    timeout: timedelta = field(default=_DEFAULT_TIMEOUT)

    #: how often the healthcheck command is retried
    retries: int = _DEFAULT_RETRIES

    @property
    def max_wait_time(self) -> timedelta:
        """The maximum time to wait until a container can become healthy"""
        return self.start_period + self.retries * self.interval + self.timeout

    @staticmethod
    def from_container_inspect(
        inspect_json: ContainerInspectHealthCheck,
    ) -> "HealthCheck":
        """Convert the json-loaded output of :command:`podman inspect $ctr` or
        :command:`docker inspect $ctr` into a :py:class:`HealthCheck`.

        """
        return HealthCheck(
            start_period=timedelta(
                microseconds=inspect_json["StartPeriod"] / 1000
            )
            if "StartPeriod" in inspect_json
            else _DEFAULT_START_PERIOD,
            interval=timedelta(microseconds=inspect_json["Interval"] / 1000)
            if "Interval" in inspect_json
            else _DEFAULT_INTERVAL,
            timeout=timedelta(microseconds=inspect_json["Timeout"] / 1000)
            if "Timeout" in inspect_json
            else _DEFAULT_TIMEOUT,
            retries=inspect_json.get("Retries", _DEFAULT_RETRIES),
        )


@dataclass(frozen=True)
class ContainerState:
    """State of the container, it is populated from the ``State`` attribute in
    the inspect of a container.

    """

    #: status of the container, e.g. ``running``, ``stopped``, etc.
    status: str
    #: True if the container is running
    running: bool
    #: True if the container has been paused
    paused: bool
    #: True if the container is restarting
    restarting: bool
    #: True if the container is has been killed by a Out Of Memory condition
    oom_killed: bool
    #: True if the container is dead
    dead: bool
    #: process id of the main container process
    pid: int
    #: status of the last health check run for this container image
    health: ContainerHealth = ContainerHealth.NO_HEALTH_CHECK


@dataclass(frozen=True)
class Config:
    """Container configuration obtained from the ``Config`` attribute in the
    inspect of a container. It features the most useful attributes and those
    that are common to both :command:`podman` and :command:`docker`.

    """

    #: User defined in the container image
    user: str

    #: true if this container has a TTY attached
    tty: bool

    #: command defined in this container
    cmd: List[str]

    #: the entrypoint of this container
    entrypoint: List[str]

    #: environment variables set in the container image
    env: Dict[str, str]

    #: name of image used to launch this container
    image: str

    #: labels of the container
    labels: Dict[str, str]

    #: Signal that will be sent to the container when it is stopped. If the
    #: container does not terminate, ``SIGKILL`` will be used afterwards.
    stop_signal: Union[int, str]

    #: The working directory of the container
    workingdir: Path

    #: optional healthcheck defined for the underlying container image
    healthcheck: Optional[HealthCheck] = None


@dataclass(frozen=True)
class ContainerNetworkSettings:
    """Network specific settings of a container."""

    #: list of ports forwarded from the container to the host
    ports: List[PortForwarding] = field(default_factory=list)

    #: IP Address of the container, if it has one
    ip_address: Optional[str] = None


@dataclass(frozen=True)
class Mount:
    """Base class for mount points"""

    #: source folder on the host (if present)
    source: str

    #: mount point in the container
    destination: str

    #: is this mount read-write?
    rw: bool


@dataclass(frozen=True)
class BindMount(Mount):
    """A bind mounted directory"""


@dataclass(frozen=True)
class VolumeMount(Mount):
    """A volume mount"""

    #: name/hash of this volume
    name: str

    #: driver that is backing this volume
    driver: str


@dataclass(frozen=True)
class ContainerInspect:
    """Common subset of the information exposed via :command:`podman inspect`
    and :command:`docker inspect`.

    """

    #: The container's ID
    id: str

    #: the container's name
    name: str

    #: program that has been launched inside the container
    path: str

    #: arguments passed to :py:attr:`path`
    args: List[str]

    #: current state of the container
    state: ContainerState

    #: hash digest of the image
    image_hash: str

    #: general configuration of the container (mostly inherited from the used image)
    config: Config

    #: Current network settings of this container
    network: ContainerNetworkSettings

    #: volumes or bind mounts mounted in this container
    mounts: List[Union[BindMount, VolumeMount]]



# pylint: disable=missing-function-docstring,missing-module-docstring
from pytest_container import GitRepositoryBuild


def test_repo_name() -> None:
    for url in ("foobar.com/repo.git", "foobar.com/repo/", "foobar.com/repo"):
        assert GitRepositoryBuild(repository_url=url).repo_name == "repo"



# pylint: disable=missing-function-docstring,missing-module-docstring
from os import remove
from time import sleep

from _pytest.config import Config
from filelock import FileLock

from pytest_container import Container
from pytest_container import DerivedContainer
from pytest_container.container import ContainerData

from .images import LEAP_URL
from .images import OPENSUSE_BUSYBOX_URL

LEAP_WITH_LOCK = DerivedContainer(
    base=OPENSUSE_BUSYBOX_URL,
    custom_entry_point="/bin/sh",
    singleton=True,
)
LEAP2_WITH_LOCK = Container(url=LEAP_URL, singleton=True)

CONTAINER_IMAGES = [LEAP_WITH_LOCK, LEAP2_WITH_LOCK]


def actual_test(container_data: ContainerData, pytestconf: Config):
    """This is the actual test function which creates a lockfile for the two
    automatic containers and keeps them around for 5 seconds. If any other test
    runs in parallel during that time, it would error out.

    """
    container_data.connection.check_output("true")
    lockfile = pytestconf.rootpath / (
        "leap.lock"
        if "leap" in container_data.image_url_or_id
        else "busybox.lock"
    )

    assert not lockfile.exists()
    with FileLock(lockfile, timeout=0):
        sleep(5)

    if lockfile.exists():
        remove(lockfile)


def test_create_conflict_1(auto_container_per_test, pytestconfig: Config):
    actual_test(auto_container_per_test, pytestconfig)


def test_create_conflict_2(auto_container_per_test, pytestconfig: Config):
    actual_test(auto_container_per_test, pytestconfig)


def test_create_conflict_3(auto_container_per_test, pytestconfig: Config):
    actual_test(auto_container_per_test, pytestconfig)


def test_create_conflict_4(auto_container_per_test, pytestconfig: Config):
    actual_test(auto_container_per_test, pytestconfig)



"""This module contains the container runtime classes abstracting away the
implementation details of container runtimes like :command:`docker` or
:command:`podman`.

"""

import json
import re
import sys
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from os import getenv
from pathlib import Path
from subprocess import check_output
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import testinfra
from _pytest.mark.structures import ParameterSet
from pytest import param

from pytest_container.inspect import BindMount
from pytest_container.inspect import Config
from pytest_container.inspect import ContainerHealth
from pytest_container.inspect import ContainerInspect
from pytest_container.inspect import ContainerNetworkSettings
from pytest_container.inspect import ContainerState
from pytest_container.inspect import HealthCheck
from pytest_container.inspect import NetworkProtocol
from pytest_container.inspect import PortForwarding
from pytest_container.inspect import VolumeMount

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

# mypy will try to import cached_property but fail to find its types
# since we run mypy with the most recent python version, we can simply import
# cached_property from stdlib and we'll be fine
if TYPE_CHECKING:  # pragma: no cover
    from functools import cached_property
else:
    try:
        from functools import cached_property
    except ImportError:
        from cached_property import cached_property

if TYPE_CHECKING:  # pragma: no cover
    import pytest_container


@dataclass(frozen=True)
class ToParamMixin:
    """
    Mixin class that gives child classes the ability to convert themselves into
    a pytest.param with self.__str__() as the default id and optional marks
    """

    marks: Any = None

    def to_pytest_param(self) -> ParameterSet:
        """Convert this class into a ``pytest.param``"""
        return param(self, id=str(self), marks=self.marks or ())


@dataclass(frozen=True)
class Version:
    """Representation of a version of the form
    ``$major.$minor.$patch[-|+]$release build $build``.

    This class supports basic comparison, e.g.:

    >>> Version(1, 0) > Version(0, 1)
    True
    >>> Version(1, 0) == Version(1, 0, 0)
    True
    >>> Version(5, 2, 6, "foobar") == Version(5, 2, 6)
    False

    Note that the patch and release fields are optional and that the release and
    build are not taken into account for less or greater than comparisons only
    for equality or inequality. I.e.:

    >>> Version(1, 0, release="16") > Version(1, 0)
    False
    >>> Version(1, 0, release="16") == Version(1, 0)
    False


    Additionally you can also pretty print it:

    >>> Version(0, 6)
    0.6
    >>> Version(0, 6, 1)
    0.6.1
    >>> Version(0, 6, 1, "asdf")
    0.6.1 build asdf

    """

    major: int = 0
    minor: int = 0
    patch: Optional[int] = None
    build: str = ""
    release: Optional[str] = None

    def __str__(self) -> str:
        return (
            f"{self.major}.{self.minor}{('.' + str(self.patch)) if self.patch is not None else ''}"
            + (f"-{self.release}" if self.release else "")
            + (f" build {self.build}" if self.build else "")
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Version):
            return False
        return (
            self.major == other.major
            and self.minor == other.minor
            and (self.patch or 0) == (other.patch or 0)
            and (self.release or "") == (other.release or "")
            and self.build == other.build
        )

    @staticmethod
    def parse(version_string: str) -> "Version":
        """Parses a version string and returns a constructed Version from that."""
        matches = re.match(
            r"(?P<major>\d+)(\.(?P<minor>\d+))?(\.(?P<patch>\d+))?"
            r"([+|-](?P<release>\S+))?( build (?P<build>\S+))?$",
            # let's first remove any leading & trailing whitespace to make our life easier
            version_string.strip(),
        )
        if not matches:
            raise ValueError(f"Invalid version string: {version_string}")

        return Version(
            major=int(matches.group("major")),
            minor=int(matches.group("minor")) if matches.group("minor") else 0,
            patch=int(matches.group("patch"))
            if matches.group("patch")
            else None,
            build=matches.group("build") or "",
            release=matches.group("release") or None,
        )

    @staticmethod
    def __generate_cmp(
        cmp_func: Callable[[int, int], bool],
    ) -> Callable[["Version", Any], bool]:
        def cmp(self: Version, other: Any) -> bool:
            if not isinstance(other, Version):
                return NotImplemented  # type: ignore

            if self.major == other.major:
                if self.minor == other.minor:
                    return cmp_func((self.patch or 0), (other.patch or 0))
                return cmp_func(self.minor, other.minor)
            return cmp_func(self.major, other.major)

        return cmp

    def __lt__(self, other: Any) -> bool:
        return Version.__generate_cmp(lambda m, n: m < n)(self, other)

    def __le__(self, other: Any) -> bool:
        return Version.__generate_cmp(lambda m, n: m <= n)(self, other)

    def __ge__(self, other: Any) -> bool:
        return Version.__generate_cmp(lambda m, n: m >= n)(self, other)

    def __gt__(self, other: Any) -> bool:
        return Version.__generate_cmp(lambda m, n: m > n)(self, other)


class OciRuntimeABC(ABC):
    """The abstract base class defining the interface of a container runtime."""

    def __init__(self, build_command: List[str], runner_binary: str) -> None:
        #: command that builds the Dockerfile in the current working directory
        self._build_command = build_command

        #: the "main" binary of this runtime, e.g. podman or docker
        self._runner_binary: str = runner_binary

    @property
    def build_command(self) -> List[str]:
        """Command that builds the :file:`Dockerfile` in the current working
        directory.

        """
        return self._build_command

    @property
    def runner_binary(self) -> str:
        """The "main" binary of this runtime, e.g. podman or docker."""
        return self._runner_binary

    def get_container_health(self, container_id: str) -> ContainerHealth:
        """Inspects the running container with the supplied id and returns its current
        health.

        """
        return self.inspect_container(container_id).state.health

    @property
    @abstractmethod
    def version(self) -> Version:
        """The version of the container runtime."""

    @abstractmethod
    def inspect_container(self, container_id: str) -> ContainerInspect:
        """Inspect the container with the provided ``container_id`` and return
        the parsed output from the container runtime as an instance of
        :py:class:`~pytest_container.inspect.ContainerInspect`.

        """

    @property
    @abstractmethod
    def supports_healthcheck_inherit_from_base(self) -> bool:
        """Indicates whether the container runtime supports that derived images
        will inherit the healthcheck from the base image.

        """


class OciRuntimeBase(OciRuntimeABC, ToParamMixin):
    """Base class of the Container Runtimes."""

    @staticmethod
    def get_image_id_from_iidfile(iidfile_path: str) -> str:
        """Returns the image id/hash from the iidfile that has been created by
        the container runtime to store the image id after a build.

        """
        with open(iidfile_path, "r", encoding="utf-8") as iidfile:
            line = iidfile.read(-1).strip().split(":")
            if len(line) == 2:
                digest_hash, digest = line
                if digest_hash != "sha256":
                    raise ValueError(f"Invalid digest hash: {digest_hash}")
                return digest
            if len(line) == 1:
                return line[0]

            raise ValueError(f"Invalid iidfile contents: {':'.join(line)}")

    def get_image_size(
        self,
        image_or_id_or_container: Union[
            str,
            "pytest_container.container.Container",
            "pytest_container.container.DerivedContainer",
        ],
    ) -> float:
        """Returns the container's size in bytes given an image id, a
        :py:class:`~pytest_container.container.Container` or a
        py:class:`~pytest_container.container.DerivedContainer`.

        """
        id_to_inspect = (
            image_or_id_or_container
            if isinstance(image_or_id_or_container, str)
            else str(image_or_id_or_container)
        )
        return float(
            check_output(
                [
                    self.runner_binary,
                    "inspect",
                    "-f",
                    '"{{ .Size }}"',
                    id_to_inspect,
                ]
            )
            .decode()
            .strip()
            .replace('"', "")
        )

    def _get_container_inspect(self, container_id: str) -> Any:
        inspect = json.loads(
            check_output([self.runner_binary, "inspect", container_id])
        )
        if len(inspect) != 1:
            raise RuntimeError(
                f"Got {len(inspect)} results back, "
                f"but expected exactly one container to match {container_id}"
            )

        return inspect[0]

    def _get_image_entrypoint_cmd(
        self, image_url_or_id: str, query_type: Literal["Entrypoint", "Cmd"]
    ) -> Optional[str]:
        """Inspect the container image with the given url or id and return its
        ``ENTRYPOINT`` or ``CMD`` or ``None`` if no entrypoint or cmd has been
        defined.

        """
        entrypoint = (
            check_output(
                [
                    self.runner_binary,
                    "inspect",
                    "-f",
                    f"{{{{.Config.{query_type}}}}}",
                    image_url_or_id,
                ]
            )
            .decode("utf-8")
            .strip()
        )
        return None if entrypoint == "[]" else entrypoint

    @staticmethod
    def _stop_signal_from_inspect_conf(inspect_conf: Any) -> Union[int, str]:
        if "StopSignal" in inspect_conf:
            raw_stop_signal = inspect_conf["StopSignal"]
            try:
                return int(raw_stop_signal)
            except ValueError:
                return str(raw_stop_signal)
        return "SIGTERM"

    @staticmethod
    def _state_from_inspect(container_inspect: Any) -> ContainerState:
        State = container_inspect["State"]
        return ContainerState(
            status=State["Status"],
            running=State["Running"],
            paused=State["Paused"],
            restarting=State["Restarting"],
            oom_killed=State["OOMKilled"],
            dead=State["Dead"],
            pid=State["Pid"],
            # depending on the podman version, this property is called either
            # Health or Healthcheck
            health=ContainerHealth(
                (State.get("Health") or State.get("Healthcheck", {})).get(
                    "Status", ""
                )
            ),
        )

    @staticmethod
    def _network_settings_from_inspect(
        container_inspect: Any,
    ) -> ContainerNetworkSettings:
        # we don't use the NetworkSettings object, but HostConfig as
        # NetworkSettings.Ports changed its structure at some point between
        # podman 1 and 4 from a dictionary into a list. However
        # HostConfig.PortBindings has always been a dictionary, so let's use
        # that for stability.
        host_config = container_inspect["HostConfig"]
        ports = []
        if "PortBindings" in host_config and host_config["PortBindings"]:
            for container_port, bindings in host_config[
                "PortBindings"
            ].items():
                if not bindings:
                    continue

                port, proto = container_port.split("/")
                # FIXME: handle multiple entries here
                ports.append(
                    PortForwarding(
                        container_port=int(port),
                        protocol=NetworkProtocol(proto),
                        host_port=int(bindings[0]["HostPort"]),
                    )
                )

        net_settings = container_inspect["NetworkSettings"]
        ip = net_settings.get("IPAddress") or None

        return ContainerNetworkSettings(ports=ports, ip_address=ip)

    @staticmethod
    def _mounts_from_inspect(
        container_inspect: Any,
    ) -> List[Union[BindMount, VolumeMount]]:
        mounts = container_inspect["Mounts"]
        res: List[Union[BindMount, VolumeMount]] = []
        for mount in mounts:
            kwargs = {
                "source": mount["Source"],
                "destination": mount["Destination"],
                "rw": mount["RW"],
            }
            if mount["Type"] == "volume":
                res.append(
                    VolumeMount(
                        name=mount["Name"], driver=mount["Driver"], **kwargs
                    )
                )
            elif mount["Type"] == "bind":
                res.append(BindMount(**kwargs))
            else:
                raise ValueError(f"Unknown mount type: {mount['Type']}")
        return res

    def __str__(self) -> str:
        return self.__class__.__name__


LOCALHOST = testinfra.host.get_host("local://")


def _get_podman_version(version_stdout: str) -> Version:
    if version_stdout[:15] != "podman version ":
        raise RuntimeError(
            f"Could not decode the podman version from '{version_stdout}'"
        )

    return Version.parse(version_stdout[15:])


def _get_buildah_version() -> Version:
    version_stdout = LOCALHOST.check_output("buildah --version")
    build_version_begin = "buildah version "
    if not version_stdout.startswith(build_version_begin):
        raise RuntimeError(
            f"Could not decode the buildah version from '{version_stdout}'"
        )

    return Version.parse(
        version_stdout.replace(build_version_begin, "").split(" ")[0]
    )


class PodmanRuntime(OciRuntimeBase):
    """The container runtime using :command:`podman` for running containers and
    :command:`buildah` for building containers.

    """

    def __init__(self) -> None:
        podman_ps = LOCALHOST.run("podman ps")
        if not podman_ps.succeeded:
            raise RuntimeError(f"`podman ps` failed with {podman_ps.stderr}")

        self._buildah_functional = LOCALHOST.run("buildah").succeeded
        super().__init__(
            build_command=(
                ["buildah", "bud", "--layers", "--force-rm"]
                if self._buildah_functional
                else ["podman", "build", "--layers", "--force-rm"]
            ),
            runner_binary="podman",
        )

    # pragma pylint: disable=used-before-assignment
    @cached_property
    def version(self) -> Version:
        """Returns the version of podman installed on the system"""
        return _get_podman_version(
            LOCALHOST.run_expect([0], "podman --version").stdout
        )

    @cached_property
    def supports_healthcheck_inherit_from_base(self) -> bool:
        # - buildah supports inheriting HEALTHCHECK since 1.25.0
        #   https://github.com/containers/buildah/blob/main/CHANGELOG.md#v1250-2022-03-25
        # - podman 4.1.0 bundles buildah >= 1.25.0
        #   https://github.com/containers/podman/blob/main/RELEASE_NOTES.md#misc-8
        podman_recent_enough = self.version >= Version(4, 1, 0)

        # if buildah isn't installed, don't check the buildah version
        if not self._buildah_functional:
            return podman_recent_enough

        return podman_recent_enough and _get_buildah_version() >= Version(
            1, 25, 0
        )

    def inspect_container(self, container_id: str) -> ContainerInspect:
        inspect = self._get_container_inspect(container_id)

        config = inspect["Config"]
        healthcheck = None
        if "Healthcheck" in config:
            healthcheck = HealthCheck.from_container_inspect(
                config["Healthcheck"]
            )

        entrypoint = config.get("Entrypoint")
        if isinstance(entrypoint, str):
            entrypoint = entrypoint.split()
        if not entrypoint:
            entrypoint = []

        conf = Config(
            user=config["User"],
            tty=config["Tty"],
            cmd=config["Cmd"],
            image=config["Image"],
            entrypoint=entrypoint,
            labels=config["Labels"],
            workingdir=Path(config["WorkingDir"]),
            env=dict([env.split("=", maxsplit=1) for env in config["Env"]]),
            stop_signal=self._stop_signal_from_inspect_conf(config),
            healthcheck=healthcheck,
        )

        state = self._state_from_inspect(inspect)

        return ContainerInspect(
            config=conf,
            state=state,
            name=inspect["Name"],
            id=inspect["Id"],
            path=inspect["Path"],
            args=inspect["Args"],
            image_hash=inspect["Image"],
            network=self._network_settings_from_inspect(inspect),
            mounts=self._mounts_from_inspect(inspect),
        )


def _get_docker_version(version_stdout: str) -> Version:
    if version_stdout[:15].lower() != "docker version ":
        raise RuntimeError(
            f"Could not decode the docker version from {version_stdout}"
        )

    return Version.parse(version_stdout[15:].replace(",", ""))


class DockerRuntime(OciRuntimeBase):
    """The container runtime using :command:`docker` for building and running
    containers."""

    def __init__(self) -> None:
        docker_ps = LOCALHOST.run("docker ps")
        if not docker_ps.succeeded:
            raise RuntimeError(f"`docker ps` failed with {docker_ps.stderr}")

        super().__init__(
            build_command=["docker", "build", "--force-rm"],
            runner_binary="docker",
        )

    @cached_property
    def version(self) -> Version:
        """Returns the version of docker installed on this system"""
        return _get_docker_version(
            LOCALHOST.run_expect([0], "docker --version").stdout
        )

    @property
    def supports_healthcheck_inherit_from_base(self) -> bool:
        return True

    def inspect_container(self, container_id: str) -> ContainerInspect:
        inspect = self._get_container_inspect(container_id)

        config = inspect["Config"]
        if config.get("Env"):
            env = dict([env.split("=", maxsplit=1) for env in config["Env"]])
        else:
            env = {}
        healthcheck = None
        if "Healthcheck" in config:
            healthcheck = HealthCheck.from_container_inspect(
                config["Healthcheck"]
            )

        conf = Config(
            user=config["User"],
            tty=config["Tty"],
            cmd=config["Cmd"],
            image=config["Image"],
            entrypoint=config["Entrypoint"],
            labels=config["Labels"],
            # docker sometimes omits the working directory,
            # then it defaults to
            workingdir=Path(config["WorkingDir"] or "/"),
            stop_signal=self._stop_signal_from_inspect_conf(config),
            env=env,
            healthcheck=healthcheck,
        )

        state = self._state_from_inspect(inspect)

        return ContainerInspect(
            config=conf,
            state=state,
            # docker prefixes the name with a / for reasons…
            name=inspect["Name"].lstrip("/"),
            id=inspect["Id"],
            path=inspect["Path"],
            args=inspect["Args"],
            image_hash=inspect["Image"],
            network=self._network_settings_from_inspect(inspect),
            mounts=self._mounts_from_inspect(inspect),
        )


def get_selected_runtime() -> OciRuntimeBase:
    """Returns the container runtime that the user selected.

    It defaults to podman and selects docker if podman & buildah are not
    present. If podman and docker are both present, then docker is returned if
    the environment variable `CONTAINER_RUNTIME` is set to `docker`.

    If neither docker nor podman are available, then a ValueError is raised.
    """
    podman_exists = LOCALHOST.exists("podman")
    docker_exists = LOCALHOST.exists("docker")

    runtime_choice = getenv("CONTAINER_RUNTIME", "podman").lower()
    if runtime_choice not in ("podman", "docker"):
        raise ValueError(f"Invalid CONTAINER_RUNTIME {runtime_choice}")

    if runtime_choice == "podman" and podman_exists:
        return PodmanRuntime()
    if runtime_choice == "docker" and docker_exists:
        return DockerRuntime()

    raise ValueError(
        "Selected runtime " + runtime_choice + " does not exist on the system"
    )



# pylint: disable=missing-function-docstring,missing-module-docstring
from pytest_container.container import Container
from pytest_container.container import ContainerData

from .images import LEAP_URL

ENV = {"SOMETHING": "42", "ANOTHER": "value", "dist": "/bin/dist"}
LEAP_WITH_ENV = Container(
    url=LEAP_URL,
    extra_environment_variables=ENV,
)


CONTAINER_IMAGES = [LEAP_WITH_ENV]


def test_environment_variables_present(auto_container: ContainerData):
    for env_var_name, env_var_val in ENV.items():
        assert (
            auto_container.connection.run_expect(
                [0], f"echo ${env_var_name}"
            ).stdout.strip()
            == env_var_val
        )



# pylint: disable=missing-function-docstring,missing-module-docstring,line-too-long
import json
from pathlib import Path
from typing import List

import pytest

from pytest_container import DerivedContainer
from pytest_container.container import ContainerData
from pytest_container.inspect import VolumeMount
from pytest_container.runtime import OciRuntimeBase
from pytest_container.runtime import PodmanRuntime

from .test_container_build import LEAP

_CTR_NAME = "foobar-12345"

IMAGE_WITH_EVERYTHING = DerivedContainer(
    singleton=True,
    extra_launch_args=["--name", _CTR_NAME],
    base=LEAP,
    containerfile="""VOLUME /src/
EXPOSE 8080 666
RUN useradd opensuse
USER opensuse
ENTRYPOINT ["/bin/bash", "-e"]
ENV HOME=/src/
WORKDIR /foobar/
ENV MY_VAR=
ENV SUFFIX_NAME=dc=example,dc=com
CMD ["/bin/sh"]
""",
)

IMAGE_WITH_STRING_CMD_AND_ENTRYPOINT = DerivedContainer(
    base=LEAP,
    containerfile="""
ENTRYPOINT /bin/bash
CMD /bin/sh
""",
)


@pytest.mark.parametrize(
    "container_per_test", [IMAGE_WITH_EVERYTHING], indirect=True
)
def test_inspect(
    container_per_test: ContainerData, container_runtime: OciRuntimeBase, host
) -> None:
    inspect = container_per_test.inspect

    assert inspect.id == container_per_test.container_id
    assert inspect.name == _CTR_NAME
    assert inspect.config.user == "opensuse"
    assert inspect.config.entrypoint == ["/bin/bash", "-e"]

    assert (
        "HOME" in inspect.config.env and inspect.config.env["HOME"] == "/src/"
    )

    # podman and docker cannot agree on what the Config.Image value is: podman
    # prefixes it with `localhost` and the full build tag
    # (i.e. `pytest_container:$digest`), while docker just uses the digest
    expected_img = (
        str(container_per_test.container)
        if container_runtime.runner_binary == "docker"
        else f"localhost/pytest_container:{container_per_test.container}"
    )

    assert inspect.config.image == expected_img
    assert inspect.config.cmd == ["/bin/sh"]
    assert Path("/foobar/") == inspect.config.workingdir

    assert (
        not inspect.state.paused
        and not inspect.state.dead
        and not inspect.state.oom_killed
        and not inspect.state.restarting
    )

    assert (
        len(inspect.mounts) == 1
        and isinstance(inspect.mounts[0], VolumeMount)
        and inspect.mounts[0].destination == "/src"
    )

    assert inspect.network.ip_address or "" == host.check_output(
        f"{container_runtime.runner_binary} inspect --format "
        '"{{ .NetworkSettings.IPAddress }}" ' + _CTR_NAME
    )


@pytest.mark.parametrize("container", [LEAP], indirect=True)
def test_inspect_unset_workdir(container: ContainerData) -> None:
    """If the container has no workdir set, check that it defaults to ``/`` as
    docker sometimes omits the workingdir setting.

    """
    assert container.inspect.config.workingdir == Path("/")


@pytest.mark.parametrize(
    "container", [IMAGE_WITH_STRING_CMD_AND_ENTRYPOINT], indirect=True
)
def test_cmd_entrypoint_parsing(container: ContainerData) -> None:
    # if only a string is added as CMD or ENTRYPOINT, then it is passed to
    # `/bin/sh -c`, hence the additional two list entries
    assert container.inspect.config.cmd == ["/bin/sh", "-c", "/bin/sh"]
    assert container.inspect.config.entrypoint == [
        "/bin/sh",
        "-c",
        "/bin/bash",
    ]


@pytest.mark.parametrize(
    "inspect_output, cmd, entrypoint",
    [
        # the outputs are the inspect of a locally build container with the
        # following containerfile:
        #
        # FROM registry.opensuse.org/opensuse/leap:15.5
        # ENTRYPOINT ["/bin/bash", "-e"]
        # CMD ["/bin/sh", "-x"]
        #
        # The first output is with podman 5, the second with podman 4.9.1
        (
            """[
     {
          "Id": "fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e",
          "Created": "2024-04-09T16:47:49.910433337+02:00",
          "Path": "/bin/bash",
          "Args": [
               "-e",
               "/bin/sh",
               "-x"
          ],
          "State": {
               "OciVersion": "1.2.0",
               "Status": "exited",
               "Running": false,
               "Paused": false,
               "Restarting": false,
               "OOMKilled": false,
               "Dead": false,
               "Pid": 0,
               "ExitCode": 126,
               "Error": "",
               "StartedAt": "2024-04-09T16:47:50.010545401+02:00",
               "FinishedAt": "2024-04-09T16:47:50.010980413+02:00",
               "CheckpointedAt": "0001-01-01T00:00:00Z",
               "RestoredAt": "0001-01-01T00:00:00Z"
          },
          "Image": "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078",
          "ImageDigest": "sha256:ff1e7475953099b8220cedba0b94cbcfe527058dbf61395d8e65b41faf98c08f",
          "ImageName": "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078",
          "Rootfs": "",
          "Pod": "",
          "ResolvConfPath": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/resolv.conf",
          "HostnamePath": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/hostname",
          "HostsPath": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/hosts",
          "StaticDir": "/home/dan/.local/share/containers/storage/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata",
          "OCIConfigPath": "/home/dan/.local/share/containers/storage/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/config.json",
          "OCIRuntime": "crun",
          "ConmonPidFile": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/conmon.pid",
          "PidFile": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/pidfile",
          "Name": "friendly_nightingale",
          "RestartCount": 0,
          "Driver": "overlay",
          "MountLabel": "system_u:object_r:container_file_t:s0:c646,c903",
          "ProcessLabel": "system_u:system_r:container_t:s0:c646,c903",
          "AppArmorProfile": "",
          "EffectiveCaps": [
               "CAP_CHOWN",
               "CAP_DAC_OVERRIDE",
               "CAP_FOWNER",
               "CAP_FSETID",
               "CAP_KILL",
               "CAP_NET_BIND_SERVICE",
               "CAP_SETFCAP",
               "CAP_SETGID",
               "CAP_SETPCAP",
               "CAP_SETUID",
               "CAP_SYS_CHROOT"
          ],
          "BoundingCaps": [
               "CAP_CHOWN",
               "CAP_DAC_OVERRIDE",
               "CAP_FOWNER",
               "CAP_FSETID",
               "CAP_KILL",
               "CAP_NET_BIND_SERVICE",
               "CAP_SETFCAP",
               "CAP_SETGID",
               "CAP_SETPCAP",
               "CAP_SETUID",
               "CAP_SYS_CHROOT"
          ],
          "ExecIDs": [],
          "GraphDriver": {
               "Name": "overlay",
               "Data": {
                    "LowerDir": "/home/dan/.local/share/containers/storage/overlay/224932ba2e71427ad30ccd681525f367b54f64745170d78444063a0a773fb0d8/diff",
                    "UpperDir": "/home/dan/.local/share/containers/storage/overlay/282deb6da22d34dfa7e5136b9690c6ea65011a91c744f9b4df838dfee38e04ae/diff",
                    "WorkDir": "/home/dan/.local/share/containers/storage/overlay/282deb6da22d34dfa7e5136b9690c6ea65011a91c744f9b4df838dfee38e04ae/work"
               }
          },
          "Mounts": [],
          "Dependencies": [],
          "NetworkSettings": {
               "EndpointID": "",
               "Gateway": "",
               "IPAddress": "",
               "IPPrefixLen": 0,
               "IPv6Gateway": "",
               "GlobalIPv6Address": "",
               "GlobalIPv6PrefixLen": 0,
               "MacAddress": "",
               "Bridge": "",
               "SandboxID": "",
               "HairpinMode": false,
               "LinkLocalIPv6Address": "",
               "LinkLocalIPv6PrefixLen": 0,
               "Ports": {},
               "SandboxKey": "",
               "Networks": {
                    "pasta": {
                         "EndpointID": "",
                         "Gateway": "",
                         "IPAddress": "",
                         "IPPrefixLen": 0,
                         "IPv6Gateway": "",
                         "GlobalIPv6Address": "",
                         "GlobalIPv6PrefixLen": 0,
                         "MacAddress": "",
                         "NetworkID": "pasta",
                         "DriverOpts": null,
                         "IPAMConfig": null,
                         "Links": null
                    }
               }
          },
          "Namespace": "",
          "IsInfra": false,
          "IsService": false,
          "KubeExitCodePropagation": "invalid",
          "lockNumber": 1,
          "Config": {
               "Hostname": "fd7fb8d8123c",
               "Domainname": "",
               "User": "",
               "AttachStdin": false,
               "AttachStdout": false,
               "AttachStderr": false,
               "Tty": false,
               "OpenStdin": false,
               "StdinOnce": false,
               "Env": [
                    "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
                    "container=podman",
                    "HOME=/root",
                    "HOSTNAME=fd7fb8d8123c"
               ],
               "Cmd": [
                    "/bin/sh",
                    "-x"
               ],
               "Image": "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078",
               "Volumes": null,
               "WorkingDir": "/",
               "Entrypoint": [
                    "/bin/bash",
                    "-e"
               ],
               "OnBuild": null,
               "Labels": {
                    "io.buildah.version": "1.35.3",
                    "org.openbuildservice.disturl": "obs://build.opensuse.org/openSUSE:Leap:15.5:Images/images/8d2ff72f5b3e4b5979c7802995bc09b7-opensuse-leap-image:docker",
                    "org.opencontainers.image.created": "2023-12-19T07:39:42.838441137Z",
                    "org.opencontainers.image.description": "Image containing a minimal environment for containers based on openSUSE Leap 15.5.",
                    "org.opencontainers.image.source": "https://build.opensuse.org/package/show/openSUSE:Leap:15.5:Images/opensuse-leap-image?rev=8d2ff72f5b3e4b5979c7802995bc09b7",
                    "org.opencontainers.image.title": "openSUSE Leap 15.5 Base Container",
                    "org.opencontainers.image.url": "https://www.opensuse.org/",
                    "org.opencontainers.image.vendor": "openSUSE Project",
                    "org.opencontainers.image.version": "15.5.5.28",
                    "org.opensuse.base.created": "2023-12-19T07:39:42.838441137Z",
                    "org.opensuse.base.description": "Image containing a minimal environment for containers based on openSUSE Leap 15.5.",
                    "org.opensuse.base.disturl": "obs://build.opensuse.org/openSUSE:Leap:15.5:Images/images/8d2ff72f5b3e4b5979c7802995bc09b7-opensuse-leap-image:docker",
                    "org.opensuse.base.reference": "registry.opensuse.org/opensuse/leap:15.5.5.28",
                    "org.opensuse.base.source": "https://build.opensuse.org/package/show/openSUSE:Leap:15.5:Images/opensuse-leap-image?rev=8d2ff72f5b3e4b5979c7802995bc09b7",
                    "org.opensuse.base.title": "openSUSE Leap 15.5 Base Container",
                    "org.opensuse.base.url": "https://www.opensuse.org/",
                    "org.opensuse.base.vendor": "openSUSE Project",
                    "org.opensuse.base.version": "15.5.5.28",
                    "org.opensuse.reference": "registry.opensuse.org/opensuse/leap:15.5.5.28"
               },
               "Annotations": {
                    "io.container.manager": "libpod",
                    "org.opencontainers.image.stopSignal": "15"
               },
               "StopSignal": "SIGTERM",
               "HealthcheckOnFailureAction": "none",
               "CreateCommand": [
                    "podman",
                    "run",
                    "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078"
               ],
               "Umask": "0022",
               "Timeout": 0,
               "StopTimeout": 10,
               "Passwd": true,
               "sdNotifyMode": "container"
          },
          "HostConfig": {
               "Binds": [],
               "CgroupManager": "systemd",
               "CgroupMode": "private",
               "ContainerIDFile": "",
               "LogConfig": {
                    "Type": "journald",
                    "Config": null,
                    "Path": "",
                    "Tag": "",
                    "Size": "0B"
               },
               "NetworkMode": "pasta",
               "PortBindings": {},
               "RestartPolicy": {
                    "Name": "",
                    "MaximumRetryCount": 0
               },
               "AutoRemove": false,
               "VolumeDriver": "",
               "VolumesFrom": null,
               "CapAdd": [],
               "CapDrop": [],
               "Dns": [],
               "DnsOptions": [],
               "DnsSearch": [],
               "ExtraHosts": [],
               "GroupAdd": [],
               "IpcMode": "shareable",
               "Cgroup": "",
               "Cgroups": "default",
               "Links": null,
               "OomScoreAdj": 0,
               "PidMode": "private",
               "Privileged": false,
               "PublishAllPorts": false,
               "ReadonlyRootfs": false,
               "SecurityOpt": [],
               "Tmpfs": {},
               "UTSMode": "private",
               "UsernsMode": "",
               "ShmSize": 65536000,
               "Runtime": "oci",
               "ConsoleSize": [
                    0,
                    0
               ],
               "Isolation": "",
               "CpuShares": 0,
               "Memory": 0,
               "NanoCpus": 0,
               "CgroupParent": "user.slice",
               "BlkioWeight": 0,
               "BlkioWeightDevice": null,
               "BlkioDeviceReadBps": null,
               "BlkioDeviceWriteBps": null,
               "BlkioDeviceReadIOps": null,
               "BlkioDeviceWriteIOps": null,
               "CpuPeriod": 0,
               "CpuQuota": 0,
               "CpuRealtimePeriod": 0,
               "CpuRealtimeRuntime": 0,
               "CpusetCpus": "",
               "CpusetMems": "",
               "Devices": [],
               "DiskQuota": 0,
               "KernelMemory": 0,
               "MemoryReservation": 0,
               "MemorySwap": 0,
               "MemorySwappiness": 0,
               "OomKillDisable": false,
               "PidsLimit": 2048,
               "Ulimits": [
                    {
                         "Name": "RLIMIT_NOFILE",
                         "Soft": 1048576,
                         "Hard": 1048576
                    },
                    {
                         "Name": "RLIMIT_NPROC",
                         "Soft": 126926,
                         "Hard": 126926
                    }
               ],
               "CpuCount": 0,
               "CpuPercent": 0,
               "IOMaximumIOps": 0,
               "IOMaximumBandwidth": 0,
               "CgroupConf": null
          }
     }
]
""",
            ["/bin/sh", "-x"],
            ["/bin/bash", "-e"],
        ),
        (
            """[
     {
          "Id": "fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e",
          "Created": "2024-04-09T16:47:49.910433337+02:00",
          "Path": "/bin/bash",
          "Args": [
               "-e",
               "/bin/sh",
               "-x"
          ],
          "State": {
               "OciVersion": "1.2.0",
               "Status": "exited",
               "Running": false,
               "Paused": false,
               "Restarting": false,
               "OOMKilled": false,
               "Dead": false,
               "Pid": 0,
               "ExitCode": 126,
               "Error": "",
               "StartedAt": "2024-04-09T16:47:50.010545401+02:00",
               "FinishedAt": "2024-04-09T16:47:50.010980413+02:00",
               "Health": {
                    "Status": "",
                    "FailingStreak": 0,
                    "Log": null
               },
               "CheckpointedAt": "0001-01-01T00:00:00Z",
               "RestoredAt": "0001-01-01T00:00:00Z"
          },
          "Image": "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078",
          "ImageDigest": "sha256:ff1e7475953099b8220cedba0b94cbcfe527058dbf61395d8e65b41faf98c08f",
          "ImageName": "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078",
          "Rootfs": "",
          "Pod": "",
          "ResolvConfPath": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/resolv.conf",
          "HostnamePath": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/hostname",
          "HostsPath": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/hosts",
          "StaticDir": "/home/dan/.local/share/containers/storage/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata",
          "OCIConfigPath": "/home/dan/.local/share/containers/storage/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/config.json",
          "OCIRuntime": "crun",
          "ConmonPidFile": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/conmon.pid",
          "PidFile": "/run/user/1000/containers/overlay-containers/fd7fb8d8123c0632558620612f900162ea71e574e3caa0ff023565dc8cd42e0e/userdata/pidfile",
          "Name": "friendly_nightingale",
          "RestartCount": 0,
          "Driver": "overlay",
          "MountLabel": "system_u:object_r:container_file_t:s0:c646,c903",
          "ProcessLabel": "system_u:system_r:container_t:s0:c646,c903",
          "AppArmorProfile": "",
          "EffectiveCaps": [
               "CAP_CHOWN",
               "CAP_DAC_OVERRIDE",
               "CAP_FOWNER",
               "CAP_FSETID",
               "CAP_KILL",
               "CAP_NET_BIND_SERVICE",
               "CAP_SETFCAP",
               "CAP_SETGID",
               "CAP_SETPCAP",
               "CAP_SETUID",
               "CAP_SYS_CHROOT"
          ],
          "BoundingCaps": [
               "CAP_CHOWN",
               "CAP_DAC_OVERRIDE",
               "CAP_FOWNER",
               "CAP_FSETID",
               "CAP_KILL",
               "CAP_NET_BIND_SERVICE",
               "CAP_SETFCAP",
               "CAP_SETGID",
               "CAP_SETPCAP",
               "CAP_SETUID",
               "CAP_SYS_CHROOT"
          ],
          "ExecIDs": [],
          "GraphDriver": {
               "Name": "overlay",
               "Data": {
                    "LowerDir": "/home/dan/.local/share/containers/storage/overlay/224932ba2e71427ad30ccd681525f367b54f64745170d78444063a0a773fb0d8/diff",
                    "UpperDir": "/home/dan/.local/share/containers/storage/overlay/282deb6da22d34dfa7e5136b9690c6ea65011a91c744f9b4df838dfee38e04ae/diff",
                    "WorkDir": "/home/dan/.local/share/containers/storage/overlay/282deb6da22d34dfa7e5136b9690c6ea65011a91c744f9b4df838dfee38e04ae/work"
               }
          },
          "Mounts": [],
          "Dependencies": [],
          "NetworkSettings": {
               "EndpointID": "",
               "Gateway": "",
               "IPAddress": "",
               "IPPrefixLen": 0,
               "IPv6Gateway": "",
               "GlobalIPv6Address": "",
               "GlobalIPv6PrefixLen": 0,
               "MacAddress": "",
               "Bridge": "",
               "SandboxID": "",
               "HairpinMode": false,
               "LinkLocalIPv6Address": "",
               "LinkLocalIPv6PrefixLen": 0,
               "Ports": {},
               "SandboxKey": "",
               "Networks": {
                    "pasta": {
                         "EndpointID": "",
                         "Gateway": "",
                         "IPAddress": "",
                         "IPPrefixLen": 0,
                         "IPv6Gateway": "",
                         "GlobalIPv6Address": "",
                         "GlobalIPv6PrefixLen": 0,
                         "MacAddress": "",
                         "NetworkID": "pasta",
                         "DriverOpts": null,
                         "IPAMConfig": null,
                         "Links": null
                    }
               }
          },
          "Namespace": "",
          "IsInfra": false,
          "IsService": false,
          "KubeExitCodePropagation": "invalid",
          "lockNumber": 1,
          "Config": {
               "Hostname": "fd7fb8d8123c",
               "Domainname": "",
               "User": "",
               "AttachStdin": false,
               "AttachStdout": false,
               "AttachStderr": false,
               "Tty": false,
               "OpenStdin": false,
               "StdinOnce": false,
               "Env": [
                    "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
                    "container=podman",
                    "HOME=/root",
                    "HOSTNAME=fd7fb8d8123c"
               ],
               "Cmd": [
                    "/bin/sh",
                    "-x"
               ],
               "Image": "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078",
               "Volumes": null,
               "WorkingDir": "/",
               "Entrypoint": "/bin/bash -e",
               "OnBuild": null,
               "Labels": {
                    "io.buildah.version": "1.35.3",
                    "org.openbuildservice.disturl": "obs://build.opensuse.org/openSUSE:Leap:15.5:Images/images/8d2ff72f5b3e4b5979c7802995bc09b7-opensuse-leap-image:docker",
                    "org.opencontainers.image.created": "2023-12-19T07:39:42.838441137Z",
                    "org.opencontainers.image.description": "Image containing a minimal environment for containers based on openSUSE Leap 15.5.",
                    "org.opencontainers.image.source": "https://build.opensuse.org/package/show/openSUSE:Leap:15.5:Images/opensuse-leap-image?rev=8d2ff72f5b3e4b5979c7802995bc09b7",
                    "org.opencontainers.image.title": "openSUSE Leap 15.5 Base Container",
                    "org.opencontainers.image.url": "https://www.opensuse.org/",
                    "org.opencontainers.image.vendor": "openSUSE Project",
                    "org.opencontainers.image.version": "15.5.5.28",
                    "org.opensuse.base.created": "2023-12-19T07:39:42.838441137Z",
                    "org.opensuse.base.description": "Image containing a minimal environment for containers based on openSUSE Leap 15.5.",
                    "org.opensuse.base.disturl": "obs://build.opensuse.org/openSUSE:Leap:15.5:Images/images/8d2ff72f5b3e4b5979c7802995bc09b7-opensuse-leap-image:docker",
                    "org.opensuse.base.reference": "registry.opensuse.org/opensuse/leap:15.5.5.28",
                    "org.opensuse.base.source": "https://build.opensuse.org/package/show/openSUSE:Leap:15.5:Images/opensuse-leap-image?rev=8d2ff72f5b3e4b5979c7802995bc09b7",
                    "org.opensuse.base.title": "openSUSE Leap 15.5 Base Container",
                    "org.opensuse.base.url": "https://www.opensuse.org/",
                    "org.opensuse.base.vendor": "openSUSE Project",
                    "org.opensuse.base.version": "15.5.5.28",
                    "org.opensuse.reference": "registry.opensuse.org/opensuse/leap:15.5.5.28"
               },
               "Annotations": {
                    "io.container.manager": "libpod",
                    "org.opencontainers.image.stopSignal": "15"
               },
               "StopSignal": 15,
               "HealthcheckOnFailureAction": "none",
               "CreateCommand": [
                    "podman",
                    "run",
                    "5a2338f9e13b00ed6ec6044a16adcc9479d7c27889ad69d5c21acb99a02bb078"
               ],
               "Umask": "0022",
               "Timeout": 0,
               "StopTimeout": 10,
               "Passwd": true,
               "sdNotifyMode": "container"
          },
          "HostConfig": {
               "Binds": [],
               "CgroupManager": "systemd",
               "CgroupMode": "private",
               "ContainerIDFile": "",
               "LogConfig": {
                    "Type": "journald",
                    "Config": null,
                    "Path": "",
                    "Tag": "",
                    "Size": "0B"
               },
               "NetworkMode": "pasta",
               "PortBindings": {},
               "RestartPolicy": {
                    "Name": "",
                    "MaximumRetryCount": 0
               },
               "AutoRemove": false,
               "VolumeDriver": "",
               "VolumesFrom": null,
               "CapAdd": [],
               "CapDrop": [],
               "Dns": [],
               "DnsOptions": [],
               "DnsSearch": [],
               "ExtraHosts": [],
               "GroupAdd": [],
               "IpcMode": "shareable",
               "Cgroup": "",
               "Cgroups": "default",
               "Links": null,
               "OomScoreAdj": 0,
               "PidMode": "private",
               "Privileged": false,
               "PublishAllPorts": false,
               "ReadonlyRootfs": false,
               "SecurityOpt": [],
               "Tmpfs": {},
               "UTSMode": "private",
               "UsernsMode": "",
               "ShmSize": 65536000,
               "Runtime": "oci",
               "ConsoleSize": [
                    0,
                    0
               ],
               "Isolation": "",
               "CpuShares": 0,
               "Memory": 0,
               "NanoCpus": 0,
               "CgroupParent": "user.slice",
               "BlkioWeight": 0,
               "BlkioWeightDevice": null,
               "BlkioDeviceReadBps": null,
               "BlkioDeviceWriteBps": null,
               "BlkioDeviceReadIOps": null,
               "BlkioDeviceWriteIOps": null,
               "CpuPeriod": 0,
               "CpuQuota": 0,
               "CpuRealtimePeriod": 0,
               "CpuRealtimeRuntime": 0,
               "CpusetCpus": "",
               "CpusetMems": "",
               "Devices": [],
               "DiskQuota": 0,
               "KernelMemory": 0,
               "MemoryReservation": 0,
               "MemorySwap": 0,
               "MemorySwappiness": 0,
               "OomKillDisable": false,
               "PidsLimit": 2048,
               "Ulimits": [
                    {
                         "Name": "RLIMIT_NOFILE",
                         "Soft": 1048576,
                         "Hard": 1048576
                    },
                    {
                         "Name": "RLIMIT_NPROC",
                         "Soft": 126926,
                         "Hard": 126926
                    }
               ],
               "CpuCount": 0,
               "CpuPercent": 0,
               "IOMaximumIOps": 0,
               "IOMaximumBandwidth": 0,
               "CgroupConf": null
          }
     }
]
""",
            ["/bin/sh", "-x"],
            ["/bin/bash", "-e"],
        ),
    ],
)
def test_podman_inspect_parsing(
    inspect_output: str,
    monkeypatch: pytest.MonkeyPatch,
    cmd: List[str],
    entrypoint: List[str],
):
    monkeypatch.setattr(
        OciRuntimeBase,
        "_get_container_inspect",
        lambda _self, _unused: json.loads(inspect_output)[0],
    )

    inspect = PodmanRuntime().inspect_container("INVALID")
    assert inspect.config.cmd == cmd
    assert inspect.config.entrypoint == entrypoint



# pylint: disable=missing-function-docstring,missing-module-docstring
from typeguard import typechecked

try:
    from typeguard.importhook import install_import_hook
except ImportError:
    from typeguard import install_import_hook

from pytest_container import add_extra_run_and_build_args_options
from pytest_container import add_logging_level_options
from pytest_container import auto_container_parametrize
from pytest_container import set_logging_level_from_cli_args


def pytest_runtest_call(item):
    # Decorate every test function [e.g. test_foo()] with typeguard's
    # typechecked() decorator.
    test_func = getattr(item, "obj", None)
    if test_func is not None:
        setattr(item, "obj", typechecked(test_func))


def pytest_generate_tests(metafunc):
    auto_container_parametrize(metafunc)


def pytest_addoption(parser):
    add_extra_run_and_build_args_options(parser)
    add_logging_level_options(parser)


def pytest_configure(config):
    set_logging_level_from_cli_args(config)
    install_import_hook("pytest_container")



# pylint: disable=missing-function-docstring,missing-module-docstring
import logging
from datetime import datetime
from datetime import timedelta
from time import sleep
from typing import Optional

import pytest

from pytest_container.container import ContainerData
from pytest_container.container import ContainerLauncher
from pytest_container.container import DerivedContainer
from pytest_container.container import ImageFormat
from pytest_container.runtime import ContainerHealth
from pytest_container.runtime import HealthCheck
from pytest_container.runtime import OciRuntimeBase
from pytest_container.runtime import get_selected_runtime

from .images import LEAP
from .images import LEAP_URL

CONTAINER_WITH_HEALTHCHECK = DerivedContainer(
    base=LEAP_URL,
    image_format=ImageFormat.DOCKER,
    # iproute2 is needed for checking the socket connection
    containerfile="""RUN zypper -n in python3 curl iproute2
EXPOSE 8000
CMD /usr/bin/python3 -c "import http.server; import os; from time import sleep; sleep(5); http.server.test(HandlerClass=http.server.SimpleHTTPRequestHandler)"
HEALTHCHECK --interval=5s --timeout=1s CMD curl --fail http://0.0.0.0:8000
""",
)

CONTAINER_DERIVING_FROM_HEALTHCHECK = DerivedContainer(
    base=CONTAINER_WITH_HEALTHCHECK, containerfile="ENV DUMMY=baz"
)


def _failing_healthcheck_container(healtcheck_args: str) -> DerivedContainer:
    return DerivedContainer(
        base=LEAP_URL,
        image_format=ImageFormat.DOCKER,
        containerfile=f"""CMD sleep 600
HEALTHCHECK {healtcheck_args} CMD false
""",
        healthcheck_timeout=timedelta(seconds=-1),
    )


CONTAINER_WITH_FAILING_HEALTHCHECK = _failing_healthcheck_container(
    "--retries=2 --interval=2s"
)

CONTAINER_THAT_FAILS_TO_LAUNCH_WITH_FAILING_HEALTHCHECK = DerivedContainer(
    base=LEAP_URL,
    image_format=ImageFormat.DOCKER,
    containerfile="""ENTRYPOINT ["/bin/false"]
HEALTHCHECK --retries=5 --timeout=10s --interval=10s CMD false
""",
)


@pytest.mark.parametrize(
    "container", [CONTAINER_WITH_HEALTHCHECK], indirect=True
)
def test_container_healthcheck(
    container: ContainerData, container_runtime: OciRuntimeBase
) -> None:
    assert (
        container_runtime.get_container_health(container.container_id)
        == ContainerHealth.HEALTHY
    )
    assert container.connection.socket("tcp://0.0.0.0:8000").is_listening


@pytest.mark.parametrize("container", [LEAP], indirect=True)
def test_container_without_healthcheck(
    container: ContainerData, container_runtime: OciRuntimeBase
) -> None:
    assert (
        container_runtime.get_container_health(container.container_id)
        == ContainerHealth.NO_HEALTH_CHECK
    )


@pytest.mark.parametrize(
    "container", [CONTAINER_WITH_FAILING_HEALTHCHECK], indirect=True
)
def test_container_with_failing_healthcheck(
    container: ContainerData, container_runtime: OciRuntimeBase
) -> None:
    # the container must be in starting state at first
    assert (
        container_runtime.get_container_health(container.container_id)
        == ContainerHealth.STARTING
    )

    # the runtime will retry the healthcheck command a few times and fail
    for _ in range(10):
        if (
            container_runtime.get_container_health(container.container_id)
            != ContainerHealth.STARTING
        ):
            break
        sleep(1)

    # the container must be unhealthy now
    assert (
        container_runtime.get_container_health(container.container_id)
        == ContainerHealth.UNHEALTHY
    )


@pytest.mark.parametrize(
    "container,healthcheck",
    [
        (
            CONTAINER_WITH_FAILING_HEALTHCHECK,
            HealthCheck(retries=2, interval=timedelta(seconds=2)),
        ),
        (
            CONTAINER_WITH_HEALTHCHECK,
            HealthCheck(
                interval=timedelta(seconds=5), timeout=timedelta(seconds=1)
            ),
        ),
        (LEAP, None),
        (
            _failing_healthcheck_container("--retries=16"),
            HealthCheck(retries=16),
        ),
        (
            _failing_healthcheck_container("--interval=21s"),
            HealthCheck(interval=timedelta(seconds=21)),
        ),
        (
            _failing_healthcheck_container("--timeout=15s"),
            HealthCheck(timeout=timedelta(seconds=15)),
        ),
        (
            _failing_healthcheck_container("--start-period=24s"),
            HealthCheck(start_period=timedelta(seconds=24)),
        ),
    ],
    indirect=["container"],
)
def test_healthcheck_timeout(
    container: ContainerData, healthcheck: Optional[HealthCheck]
) -> None:
    assert container.inspect.config.healthcheck == healthcheck


@pytest.mark.skipif(
    not get_selected_runtime().supports_healthcheck_inherit_from_base,
    reason="Runtime does not inheriting HEALTHCHECK from base images",
)
@pytest.mark.parametrize(
    "container", [CONTAINER_DERIVING_FROM_HEALTHCHECK], indirect=True
)
def test_image_deriving_from_healthcheck_has_healthcheck(
    container: ContainerData, container_runtime: OciRuntimeBase
) -> None:
    assert (
        container_runtime.get_container_health(container.container_id)
        == ContainerHealth.HEALTHY
    )


def test_container_that_doesnt_run_is_reported_unhealthy(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    before = datetime.now()
    with pytest.raises(RuntimeError) as rt_err_ctx:
        with ContainerLauncher(
            container=CONTAINER_THAT_FAILS_TO_LAUNCH_WITH_FAILING_HEALTHCHECK,
            container_runtime=container_runtime,
            rootdir=pytestconfig.rootpath,
        ) as launcher:
            launcher.launch_container()
            assert False, "The container must fail to launch"
    after = datetime.now()

    time_to_fail = after - before
    assert time_to_fail < timedelta(seconds=15), (
        f"container must fail quickly (threshold 15s), but it took {time_to_fail.total_seconds()}"
    )
    assert "not running, got " in str(rt_err_ctx.value)


def test_container_launcher_logs_correct_healthcheck_timeout(
    container_runtime: OciRuntimeBase,
    pytestconfig: pytest.Config,
    caplog: pytest.LogCaptureFixture,
) -> None:
    caplog.set_level(logging.DEBUG)
    ctr = DerivedContainer(
        base=LEAP_URL,
        image_format=ImageFormat.DOCKER,
        containerfile="HEALTHCHECK --retries=5 --timeout=10s --interval=10s CMD true",
    )
    with ContainerLauncher(
        container=ctr,
        container_runtime=container_runtime,
        rootdir=pytestconfig.rootpath,
    ) as launcher:
        launcher.launch_container()
        assert launcher.container_data.inspect.config.healthcheck
        timeout = (
            launcher.container_data.inspect.config.healthcheck.max_wait_time
        )
        assert timeout == timedelta(seconds=60)

    assert (
        "Container has a healthcheck defined, will wait at most 60.0 s"
        in caplog.text
    )



# pylint: disable=missing-function-docstring,missing-module-docstring
from pathlib import Path
from typing import Union

import pytest
from _pytest.mark import ParameterSet

from pytest_container import DerivedContainer
from pytest_container import MultiStageBuild
from pytest_container import OciRuntimeBase
from pytest_container import container_from_pytest_param
from pytest_container import container_to_pytest_param
from pytest_container import get_extra_build_args
from pytest_container.container import ContainerData
from pytest_container.container import container_and_marks_from_pytest_param
from pytest_container.pod import Pod
from pytest_container.pod import pod_from_pytest_param

from .images import LEAP
from .images import TEST_POD

LEAP_PARAM = pytest.param(LEAP)
LEAP_PARAM_2 = pytest.param(
    LEAP, marks=pytest.mark.skipif(False, reason="Don't skip this please :)")
)

TEMPLATE = """FROM $builder as builder
WORKDIR /src
RUN touch test.sh && chmod +x test.sh

FROM $runner as runner
WORKDIR /bin
COPY --from=builder /src/test.sh .
ENTRYPOINT ["/bin/test.sh"]
"""

CONTAINER_IMAGES = [LEAP_PARAM]


def test_multistage_with_param(
    tmp_path: Path,
    pytestconfig: pytest.Config,
    container_runtime: OciRuntimeBase,
):
    MultiStageBuild(
        containers={"builder": LEAP_PARAM, "runner": LEAP_PARAM_2},
        containerfile_template=TEMPLATE,
    ).build(
        tmp_path,
        pytestconfig.rootpath,
        container_runtime,
        extra_build_args=get_extra_build_args(pytestconfig),
    )


def test_multistage_build_invalid_param(
    tmp_path: Path,
    pytestconfig: pytest.Config,
    container_runtime: OciRuntimeBase,
):
    with pytest.raises(ValueError):
        MultiStageBuild(
            containers={"runner": pytest.param()},
            containerfile_template=TEMPLATE,
        ).prepare_build(tmp_path, container_runtime, pytestconfig.rootpath)

    with pytest.raises(ValueError):
        MultiStageBuild(
            containers={"runner": pytest.param(1.0)},
            containerfile_template=TEMPLATE,
        ).prepare_build(tmp_path, container_runtime, pytestconfig.rootpath)


@pytest.mark.parametrize(
    "container_per_test", [container_to_pytest_param(LEAP)], indirect=True
)
def test_container_build_with_param(container_per_test: ContainerData):
    container_per_test.connection.run_expect([0], "true")


def test_auto_container_build_with_param(auto_container: ContainerData):
    auto_container.connection.run_expect([0], "true")


def test_container_to_pytest_param() -> None:
    param = container_to_pytest_param(LEAP)
    assert len(param.values) == 1 and param.values[0] == LEAP
    assert param.id == str(LEAP)
    assert len(param.marks) == 0

    skip_mark = pytest.mark.skip()
    param_with_marks = container_to_pytest_param(LEAP, marks=skip_mark)
    assert (
        len(param_with_marks.marks) == 1
        and param_with_marks.marks[0] == skip_mark
    )

    param_with_marks_2 = container_to_pytest_param(LEAP, marks=[skip_mark])
    assert (
        len(param_with_marks_2.marks) == 1
        and param_with_marks_2.marks[0] == skip_mark
    )


def test_container_from_pytest_param() -> None:
    assert container_from_pytest_param(container_to_pytest_param(LEAP)) == LEAP
    assert container_from_pytest_param(pytest.param(LEAP, 1, "a")) == LEAP
    assert container_from_pytest_param(LEAP) == LEAP

    derived = DerivedContainer(base=LEAP, containerfile="ENV foo=bar")
    assert (
        container_from_pytest_param(container_to_pytest_param(derived))
        == derived
    )
    assert container_from_pytest_param(derived) == derived

    with pytest.raises(ValueError) as val_err_ctx:
        container_from_pytest_param(pytest.param(16, 45))
    assert "Invalid pytest.param values" in str(val_err_ctx.value)
    assert "(16, 45)" in str(val_err_ctx.value)


def test_container_and_marks_from_pytest_param() -> None:
    cont, marks = container_and_marks_from_pytest_param(
        container_to_pytest_param(LEAP)
    )
    assert cont == LEAP and not marks

    cont, marks = container_and_marks_from_pytest_param(
        pytest.param(LEAP, 1, "a")
    )
    assert cont == LEAP and not marks

    assert container_and_marks_from_pytest_param(LEAP) == (LEAP, None)

    derived = DerivedContainer(base=LEAP, containerfile="ENV foo=bar")
    cont, marks = container_and_marks_from_pytest_param(
        container_to_pytest_param(derived)
    )
    assert cont == derived and not marks

    assert container_and_marks_from_pytest_param(derived) == (derived, None)

    with pytest.raises(ValueError) as val_err_ctx:
        container_and_marks_from_pytest_param(pytest.param(16, 45))
    assert "Invalid pytest.param values" in str(val_err_ctx.value)
    assert "(16, 45)" in str(val_err_ctx.value)


@pytest.mark.parametrize(
    "param,expected_pod",
    [(TEST_POD, TEST_POD), (pytest.param(TEST_POD), TEST_POD)],
)
def test_pod_from_pytest_param(
    param: Union[Pod, ParameterSet], expected_pod: Pod
) -> None:
    assert pod_from_pytest_param(param) == expected_pod


def test_invalid_pod_from_pytest_param() -> None:
    with pytest.raises(ValueError) as val_err_ctx:
        pod_from_pytest_param(pytest.param([1, 2]))

    assert "Invalid pytest.param" in str(val_err_ctx.value)



# pylint: disable=missing-function-docstring,missing-module-docstring,line-too-long,trailing-whitespace
from pathlib import Path

import pytest

from pytest_container.container import DerivedContainer
from pytest_container.container import ImageFormat
from pytest_container.container import PortForwarding
from pytest_container.pod import Pod
from pytest_container.pod import PodData
from pytest_container.pod import PodLauncher
from pytest_container.pod import infra_container_id_from_pod_inspect
from pytest_container.runtime import OciRuntimeBase
from pytest_container.runtime import PodmanRuntime

from .images import BUSYBOX
from .images import CONTAINER_THAT_FAILS_TO_LAUNCH
from .images import LEAP
from .images import LEAP_WITH_MAN
from .images import NGINX_URL
from .images import TEST_POD
from .images import WEB_SERVER

TEST_POD_WITHOUT_PORTS = Pod(containers=[LEAP, LEAP_WITH_MAN, BUSYBOX])


NGINX_PROXY = DerivedContainer(
    base=NGINX_URL,
    containerfile=r"""COPY tests/files/nginx.conf /etc/nginx/nginx.conf
RUN set -eu; zypper -n ref; zypper -n in netcat; zypper -n clean; rm -rf /var/log/{zypp,zypper.log,lastlog}

HEALTHCHECK --interval=5s --timeout=1s CMD echo "GET /" | nc localhost 80
""",
    image_format=ImageFormat.DOCKER,
)

PROXY_POD = Pod(
    containers=[WEB_SERVER, NGINX_PROXY],
    forwarded_ports=[PortForwarding(container_port=80)],
)


def test_pod_launcher(
    tmp_path: Path,
    container_runtime: OciRuntimeBase,
    pytestconfig: pytest.Config,
) -> None:
    if container_runtime != PodmanRuntime():
        pytest.skip("pods only work with podman")

    pidfile_path = str(tmp_path / "pidfile")
    with PodLauncher(
        pod=TEST_POD,
        rootdir=pytestconfig.rootpath,
        extra_pod_create_args=["--pod-id-file", pidfile_path],
    ) as launcher:
        launcher.launch_pod()
        pod_data = launcher.pod_data
        assert pod_data.pod_id and pod_data.infra_container_id

        with open(pidfile_path, encoding="utf-8") as pyproject:
            assert pod_data.pod_id == pyproject.read()

        assert (
            len(pod_data.forwarded_ports) == 2
            and pod_data.forwarded_ports[0].container_port == 80
            and pod_data.forwarded_ports[1].container_port == 22
        )
        assert (
            len(pod_data.pod.containers) == 3
            and len(pod_data.container_data) == 3
        )


def test_pod_launcher_pod_data_not_ready(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config
) -> None:
    if container_runtime != PodmanRuntime():
        pytest.skip("pods only work with podman")

    with PodLauncher.from_pytestconfig(
        pod=TEST_POD, pytestconfig=pytestconfig
    ) as launcher:
        with pytest.raises(RuntimeError) as rt_err_ctx:
            _ = launcher.pod_data

        assert "Pod has not been created" in str(rt_err_ctx.value)

        launcher.launch_pod()
        assert launcher.pod_data


def test_pod_launcher_cleanup(
    container_runtime: OciRuntimeBase, pytestconfig: pytest.Config, host
) -> None:
    if container_runtime != PodmanRuntime():
        pytest.skip("pods only work with podman")

    name = "i_will_fail_to_launch"

    with pytest.raises(RuntimeError) as rt_err_ctx:
        with PodLauncher.from_pytestconfig(
            pod=Pod(containers=[LEAP, CONTAINER_THAT_FAILS_TO_LAUNCH]),
            pytestconfig=pytestconfig,
            pod_name=name,
        ) as launcher:
            launcher.launch_pod()
            assert False, "This code must be unreachable"

    assert "did not become healthy" in str(rt_err_ctx.value)

    # the pod should be gone
    assert (
        name
        in host.run_expect([125], f"podman pod inspect {name}").stderr.strip()
    )


def test_pod_launcher_fails_with_non_podman(
    container_runtime: OciRuntimeBase,
) -> None:
    if container_runtime == PodmanRuntime():
        pytest.skip("pods work with podman")

    with pytest.raises(RuntimeError) as rt_err_ctx:
        with PodLauncher(pod=TEST_POD, rootdir=Path("/tmp")) as _:
            pass

    assert "pods can only be created with podman" in str(rt_err_ctx.value)


@pytest.mark.parametrize("pod_per_test", [PROXY_POD], indirect=True)
def test_proxy_pod(pod_per_test: PodData, host) -> None:
    assert (
        pod_per_test.forwarded_ports and len(pod_per_test.forwarded_ports) == 1
    )
    assert host.socket(
        f"tcp://0.0.0.0:{pod_per_test.forwarded_ports[0].host_port}"
    ).is_listening

    assert (
        "Hello Green World"
        in host.run_expect(
            [0],
            f"curl --fail http://0.0.0.0:{pod_per_test.forwarded_ports[0].host_port}",
        ).stdout
    )


@pytest.mark.parametrize(
    "pod", [TEST_POD, TEST_POD_WITHOUT_PORTS], indirect=True
)
def test_pod_fixture(pod: PodData) -> None:
    assert pod.pod_id

    for cont_data in pod.container_data:
        # leap has /etc/os-release
        if cont_data.connection.file("/etc/os-release").exists:
            assert (
                "leap"
                in cont_data.connection.run_expect([0], "cat /etc/os-release")
                .stdout.strip()
                .lower()
            )
        # busybox doesn't, but it has /bin/busybox ;-)
        else:
            assert cont_data.connection.file("/bin/busybox").exists


def test_launcher_pod_data_uninitialized() -> None:
    with pytest.raises(RuntimeError) as rt_ctx:
        _ = PodLauncher(TEST_POD, Path("/")).pod_data

    assert "Pod has not been created" in str(rt_ctx.value)


@pytest.mark.parametrize(
    "inspect_output, expected_id",
    (
        # podman 1.6.4 (CentOS 7)
        (
            """{
     "Config": {
          "id": "2df19026a989667f0c9956723a45b25ccd32c0efaec19d375727d37622709040",
          "name": "happy_beaver",
          "hostname": "happy_beaver",
          "labels": {
"""
            # This whitespace is present in the output and should stay there,
            # but ruff complains about it then if it's a single multiline
            # string. But if we split it up, everything is fine ¯\_(ツ)_/¯
            + "               "
            + """
          },
          "cgroupParent": "machine.slice",
          "sharesCgroup": true,
          "sharesIpc": true,
          "sharesNet": true,
          "sharesUts": true,
          "infraConfig": {
               "makeInfraContainer": true,
               "infraPortBindings": null
          },
          "created": "2024-04-09T12:56:00.106588707Z",
          "lockID": 0
     },
     "State": {
          "cgroupPath": "machine.slice/machine-libpod_pod_2df19026a989667f0c9956723a45b25ccd32c0efaec19d375727d37622709040.slice",
          "infraContainerID": "9c77fc16ce2cd7b7b648ac161596b8e2c4b66e6b388b1262de402bb76b2563a4"
     },
     "Containers": [
          {
               "id": "9c77fc16ce2cd7b7b648ac161596b8e2c4b66e6b388b1262de402bb76b2563a4",
               "state": "configured"
          }
     ]
}
""",
            "9c77fc16ce2cd7b7b648ac161596b8e2c4b66e6b388b1262de402bb76b2563a4",
        ),
        # podman < 5
        (
            """{
     "Id": "759b48e641608d2dc8ecef44084d4671b9d127e4f39d9cdfa9c9bb326cfe64f0",
     "Name": "git",
     "Created": "2024-04-09T14:34:44.307240325+02:00",
     "CreateCommand": [
          "podman",
          "pod",
          "create",
          "git"
     ],
     "ExitPolicy": "continue",
     "State": "Created",
     "Hostname": "",
     "CreateCgroup": true,
     "CgroupParent": "user.slice",
     "CgroupPath": "user.slice/user-1000.slice/user@1000.service/user.slice/user-libpod_pod_759b48e641608d2dc8ecef44084d4671b9d127e4f39d9cdfa9c9bb326cfe64f0.slice",
     "CreateInfra": true,
     "InfraContainerID": "b4541110f6c7f06aec932b1e5381682df306d52b6eb080d44aa09c2865469584",
     "InfraConfig": {
          "PortBindings": {},
          "HostNetwork": false,
          "StaticIP": "",
          "StaticMAC": "",
          "NoManageResolvConf": false,
          "DNSServer": null,
          "DNSSearch": null,
          "DNSOption": null,
          "NoManageHosts": false,
          "HostAdd": null,
          "Networks": null,
          "NetworkOptions": null,
          "pid_ns": "private",
          "userns": "host",
          "uts_ns": "private"
     },
     "SharedNamespaces": [
          "ipc",
          "net",
          "uts"
     ],
     "NumContainers": 1,
     "Containers": [
          {
               "Id": "b4541110f6c7f06aec932b1e5381682df306d52b6eb080d44aa09c2865469584",
               "Name": "759b48e64160-infra",
               "State": "created"
          }
     ],
     "LockNumber": 0
}
""",
            "b4541110f6c7f06aec932b1e5381682df306d52b6eb080d44aa09c2865469584",
        ),  # podman > 5
        (
            """[
     {
          "Id": "759b48e641608d2dc8ecef44084d4671b9d127e4f39d9cdfa9c9bb326cfe64f0",
          "Name": "git",
          "Created": "2024-04-09T14:34:44.307240325+02:00",
          "CreateCommand": [
               "podman",
               "pod",
               "create",
               "git"
          ],
          "ExitPolicy": "continue",
          "State": "Created",
          "Hostname": "",
          "CreateCgroup": true,
          "CgroupParent": "user.slice",
          "CgroupPath": "user.slice/user-1000.slice/user@1000.service/user.slice/user-libpod_pod_759b48e641608d2dc8ecef44084d4671b9d127e4f39d9cdfa9c9bb326cfe64f0.slice",
          "CreateInfra": true,
          "InfraContainerID": "b4541110f6c7f06aec932b1e5381682df306d52b6eb080d44aa09c2865469584",
          "InfraConfig": {
               "PortBindings": {},
               "HostNetwork": false,
               "StaticIP": "",
               "StaticMAC": "",
               "NoManageResolvConf": false,
               "DNSServer": null,
               "DNSSearch": null,
               "DNSOption": null,
               "NoManageHosts": false,
               "HostAdd": null,
               "Networks": null,
               "NetworkOptions": null,
               "pid_ns": "private",
               "userns": "host",
               "uts_ns": "private"
          },
          "SharedNamespaces": [
               "net",
               "uts",
               "ipc"
          ],
          "NumContainers": 1,
          "Containers": [
               {
                    "Id": "b4541110f6c7f06aec932b1e5381682df306d52b6eb080d44aa09c2865469584",
                    "Name": "759b48e64160-infra",
                    "State": "created"
               }
          ],
          "LockNumber": 0
     }
]
""",
            "b4541110f6c7f06aec932b1e5381682df306d52b6eb080d44aa09c2865469584",
        ),
    ),
)
def test_get_infra_container_id(inspect_output: str, expected_id: str) -> None:
    assert (
        infra_container_id_from_pod_inspect(inspect_output.encode())
        == expected_id
    )
